import random
from typing import List

import numpy as np

from ..base import BaseVideoPreprocessor, RawChunkPreprocessor
from ..helper import split_video_frames, phash_descriptor
from ..io_utils import video as video_util
from ...proto import gnes_pb2, array2blob, blob2array


class FFmpegPreprocessor(BaseVideoPreprocessor):

    def __init__(self,
                 frame_size: str = '192:168',
                 frame_rate: int = 10,
                 frame_num: int = -1,
                 duplicate_rm: bool = True,
                 use_phash_weight: bool = False,
                 phash_thresh: int = 5,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.frame_size = frame_size
        self.frame_rate = frame_rate
        self.frame_num = frame_num
        self.phash_thresh = phash_thresh
        self.duplicate_rm = duplicate_rm
        self.use_phash_weight = use_phash_weight

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)

        # video could't be processed from ndarray!
        # only bytes can be passed into ffmpeg pipeline
        if doc.raw_bytes:
            frames = video_util.capture_frames(input_data=doc.raw_bytes, scale=self.frame_size,
                                               fps=self.frame_rate, vframes=self.frame_num)
            # remove dupliated key frames by phash value
            if self.duplicate_rm:
                frames = self.duplicate_rm_hash(frames)

            if self.use_phash_weight:
                weight = FFmpegPreprocessor.pic_weight(frames)
            else:
                weight = [1 / len(frames)] * len(frames)

            for ci, chunk in enumerate(frames):
                c = doc.chunks.add()
                c.doc_id = doc.doc_id
                c.blob.CopyFrom(array2blob(chunk))
                c.offset = ci
                c.weight = weight[ci]

        else:
            self.logger.error('bad document: "raw_bytes" is empty!')

    @staticmethod
    def pic_weight(images: List['np.ndarray']) -> List[float]:
        import cv2
        weight = np.zeros([len(images)])
        # n_channel is usually 3 for RGB images
        n_channel = images[0].shape[-1]
        for i, image in enumerate(images):
            weight[i] = sum([
                cv2.calcHist([image], [_], None, [256], [0, 256]).var()
                for _ in range(n_channel)
            ])
        weight = weight / weight.sum()

        # normalized result
        weight = np.exp(-weight * 10)
        return weight / weight.sum()

    def duplicate_rm_hash(self,
                          images: List['np.ndarray']) -> List['np.ndarray']:
        hash_list = [phash_descriptor(_) for _ in images]
        ret = []
        for i, h in enumerate(hash_list):
            flag = 1
            if len(ret) >= 1:
                # only keep images with high phash diff
                # comparing only last kept 9 pics
                for j in range(1, min(len(ret) + 1, 9)):
                    dist = abs(ret[-j][1] - h)
                    if dist < self.phash_thresh:
                        flag = 0
                        break

            if flag:
                ret.append((i, h))

        return [images[_[0]] for _ in ret]


class FFmpegVideoSegmentor(BaseVideoPreprocessor):
    def __init__(self,
                 frame_size: str = '192:168',
                 frame_rate: int = 10,
                 frame_num: int = -1,
                 segment_method: str = 'cut_by_frame',
                 segment_interval: int = -1,
                 segment_num: int = 3,
                 max_frames_per_doc: int = -1,
                 use_image_input: bool = False,
                 splitter: str = '__split__',
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.frame_size = frame_size
        self.frame_rate = frame_rate
        self.frame_num = frame_num
        self.segment_method = segment_method
        self.segment_interval = segment_interval
        self.segment_num = segment_num
        self.max_frames_per_doc = max_frames_per_doc
        self.splitter = splitter
        self.use_image_input = use_image_input

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)
        from sklearn.cluster import KMeans
        if doc.raw_bytes:
            if self.use_image_input:
                frames = split_video_frames(doc.raw_bytes, self.splitter)
            else:
                frames = video_util.capture_frames(input_data=doc.raw_bytes, scale=self.frame_size,
                                                   fps=self.frame_rate, vframes=self.frame_num)
            if self.max_frames_per_doc > 0:
                random_id = random.sample(range(len(frames)),
                                          k=min(self.max_frames_per_doc, len(frames)))
                frames = [frames[i] for i in sorted(random_id)]

            sub_videos = []
            if len(frames) >= 1:
                # cut by frame: should specify how many frames to cut
                if self.segment_method == 'cut_by_frame':
                    if self.segment_interval == -1:
                        sub_videos = [frames]
                    else:
                        sub_videos = [frames[_: _ + self.segment_interval]
                                      for _ in range(0, len(frames), self.segment_interval)]
                # cut by num: should specify how many chunks for each doc
                elif self.segment_method == 'cut_by_num':
                    if self.segment_num >= 2:
                        _interval = len(frames) // (self.segment_num - 1)
                        sub_videos = [frames[_: _ + _interval]
                                      for _ in range(0, len(frames), _interval)]
                    else:
                        sub_videos = [frames]

                # cut by clustering: params required
                #   segment_num
                elif self.segment_method == 'cut_by_clustering':
                    if self.segment_num >= 2:
                        hash_v = [phash_descriptor(_).hash for _ in frames]
                        hash_v = np.array(hash_v, dtype=np.int32).reshape([len(hash_v), -1])
                        label_v = KMeans(n_clusters=self.segment_num).fit_predict(hash_v)
                        sub_videos = [[frames[i] for i, j in enumerate(label_v) if j == _] for _ in
                                      range(self.segment_num)]
                    else:
                        sub_videos = [frames]

                for ci, chunk in enumerate(sub_videos):
                    c = doc.chunks.add()
                    c.doc_id = doc.doc_id
                    c.blob.CopyFrom(array2blob(np.array(chunk, dtype=np.uint8)))
                    c.offset = ci
                    c.weight = 1 / len(sub_videos)

            else:
                self.logger.warning('bad document: no key frames extracted')
        else:
            self.logger.error('bad document: "raw_bytes" is empty!')


class GifChunkPreprocessor(RawChunkPreprocessor, BaseVideoPreprocessor):
    @staticmethod
    def _parse_chunk(chunk: 'gnes_pb2.Chunk', *args, **kwargs):
        from ..io_utils import gif as gif_util

        return gif_util.encode_video(blob2array(chunk.blob), frame_rate=10)



from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.parse import parse_qs

from .base import YamlComposer, parse_http_data
from ..cli.parser import set_composer_parser
from ..helper import set_logger


class YamlComposerHttp:
    def __init__(self, args):
        self.args = args
        self.logger = set_logger(self.__class__.__name__, self.args.verbose)

    class _HttpServer(BaseHTTPRequestHandler):
        args = set_composer_parser().parse_args([])
        default_html = YamlComposer(args).build_all()['html']

        def _set_response(self, msg: str, code: int = 200):
            self.send_response(code)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            self.wfile.write(msg.encode('utf-8'))

        def do_GET(self):
            if str(self.path) != '/':
                self._set_response('<h1>"%s" is not a valid entrypoint</h1>' % self.path, 400)
                return
            self._set_response(self.default_html)

        def do_POST(self):
            if str(self.path) != '/generate':
                self._set_response('<h1>"%s" is not a valid entrypoint</h1>' % self.path, 400)
                return
            content_length = int(self.headers['Content-Length'])  # <--- Gets the size of data
            data = self.rfile.read(content_length)  # <--- Gets the data itself

            data = {k: v[0] for k, v in parse_qs(data.decode('utf-8')).items()}
            self._set_response(*parse_http_data(data, self.args))

    def run(self):
        httpd = HTTPServer(('0.0.0.0', self.args.http_port), self._HttpServer)
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            pass
        finally:
            httpd.server_close()



from ..base import BaseVideoPreprocessor
from ..io_utils import video, gif, webp
from ...proto import gnes_pb2, blob2array


class VideoEncoderPreprocessor(BaseVideoPreprocessor):
    def __init__(self, frame_rate: int = 10, pix_fmt: str = 'rgb24', video_format: str = "mp4", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pix_fmt = pix_fmt
        self.frame_rate = frame_rate
        self.video_format = video_format

        if self.video_format not in ['mp4', 'gif', 'webp']:
            raise ValueError("%s encoder has not been supported!" % (self.video_format))

    def _encode(self, images: 'np.ndarray'):
        encoder = None
        if self.video_format == 'mp4':
            encoder = video
        elif self.video_format == 'gif':
            encoder = gif
        elif self.video_format == 'webp':
            encoder = webp

        return encoder.encode_video(images, pix_fmt=self.pix_fmt, frame_rate=self.frame_rate)

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)
        if len(doc.chunks) > 0:
            for chunk in doc.chunks:
                images = blob2array(chunk.blob)
                chunk.raw = self._encode(images)
        elif doc.WhichOneof('raw_data'):
            raw_type = type(getattr(doc, doc.WhichOneof('raw_data')))
            if raw_type == gnes_pb2.NdArray:
                images = blob2array(doc.raw_video)
                doc.raw_bytes = self._encode(images)
            else:
                self.logger.error('bad document: "doc.raw_video" is empty!')
        else:
            self.logger.error(
                'bad document: "doc.chunks" and "doc.raw_video" is empty!')



from typing import List

import math
import numpy as np

from gnes.preprocessor.base import BaseVideoPreprocessor
from gnes.preprocessor.helper import compute_descriptor, compare_descriptor, detect_peak_boundary, compare_ecr
from gnes.preprocessor.io_utils import video
from gnes.proto import gnes_pb2, array2blob, blob2array


class ShotDetectorPreprocessor(BaseVideoPreprocessor):
    store_args_kwargs = True

    def __init__(self,
                 descriptor: str = 'block_hsv_histogram',
                 distance_metric: str = 'bhattacharya',
                 detect_method: str = 'threshold',
                 frame_size: str = None,
                 frame_rate: int = 10,
                 vframes: int = -1,
                 sframes: int = -1,
                 drop_raw_data: bool = False,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.frame_size = frame_size
        self.descriptor = descriptor
        self.distance_metric = distance_metric
        self.detect_method = detect_method
        self.frame_rate = frame_rate
        self.vframes = vframes
        self.sframes = sframes
        self.drop_raw_data = drop_raw_data
        self._detector_kwargs = kwargs

    def detect_shots(self, frames: 'np.ndarray') -> List[List['np.ndarray']]:
        descriptors = []
        for frame in frames:
            descriptor = compute_descriptor(
                frame, method=self.descriptor, **self._detector_kwargs)
            descriptors.append(descriptor)

        # compute distances between frames
        if self.distance_metric == 'edge_change_ration':
            dists = compare_ecr(descriptors, **self._detector_kwargs)
        else:
            dists = [
                compare_descriptor(pair[0], pair[1], self.distance_metric)
                for pair in zip(descriptors[:-1], descriptors[1:])
            ]
            self._detector_kwargs['neigh_avg'] = 0

        shot_bounds = detect_peak_boundary(dists, self.detect_method, **self._detector_kwargs)

        shots = []
        for ci in range(0, len(shot_bounds) - 1):
            shots.append(frames[shot_bounds[ci]:shot_bounds[ci + 1]].copy())

        return shots

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)

        video_frames = []

        if doc.WhichOneof('raw_data'):
            raw_type = type(getattr(doc, doc.WhichOneof('raw_data')))
            if doc.raw_bytes:
                video_frames = video.capture_frames(
                    input_data=doc.raw_bytes,
                    scale=self.frame_size,
                    fps=self.frame_rate,
                    vframes=self.vframes)
            elif raw_type == gnes_pb2.NdArray:
                video_frames = blob2array(doc.raw_video)
                if self.vframes > 0:
                    video_frames = video_frames[0:self.vframes, :].copy()

            num_frames = len(video_frames)
            if num_frames > 0:
                shots = self.detect_shots(video_frames)
                for ci, frames in enumerate(shots):
                    c = doc.chunks.add()
                    c.doc_id = doc.doc_id
                    c.offset = ci
                    shot_len = len(frames)
                    c.weight = shot_len / num_frames
                    if self.sframes > 0 and shot_len > self.sframes:
                        if shot_len >= 2 * self.sframes:
                            step = math.ceil(shot_len / self.sframes)
                            frames = frames[::step]
                        else:
                            idx = np.sort(np.random.choice(shot_len, self.sframes, replace=False))
                            frames = [frames[idx_] for idx_ in idx]

                    chunk_data = np.array(frames)
                    c.blob.CopyFrom(array2blob(chunk_data))
            else:
                self.logger.error(
                    'bad document: "raw_bytes" or "raw_video" is empty!')
        else:
            self.logger.error('bad document: "raw_data" is empty!')

        if self.drop_raw_data:
            self.logger.info("document raw data will be cleaned!")
            doc.ClearField('raw_data')



import json
import re
import string

from ..base import BaseTextPreprocessor
from ...proto import gnes_pb2


class SentSplitPreprocessor(BaseTextPreprocessor):
    def __init__(self,
                 min_sent_len: int = 1,
                 max_sent_len: int = 256,
                 deliminator: str = '.!?。！？',
                 is_json: bool = False,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.min_sent_len = min_sent_len
        self.max_sent_len = max_sent_len
        self.deliminator = deliminator
        self.is_json = is_json

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)
        d = doc.raw_bytes.decode()
        if self.is_json:
            d = json.loads(d)
            doc.raw_text = d.pop('Content')
            doc.meta_info = json.dumps(d).encode()
        else:
            doc.raw_text = d

        ret = [(m.group(0), m.start(), m.end()) for m in
               re.finditer(r'[^{0}]+[{0}]'.format(self.deliminator), doc.raw_text)]
        if not ret:
            ret = [(doc.raw_text, 0, len(doc.raw_text))]
        for ci, (r, s, e) in enumerate(ret):
            f = ''.join(filter(lambda x: x in string.printable, r))
            f = re.sub('\n+', ' ', f).strip()
            if len(f) > self.min_sent_len:
                c = doc.chunks.add()
                c.doc_id = doc.doc_id
                c.text = f[:self.max_sent_len]
                c.offset = ci
                c.weight = len(c.text) / len(doc.raw_text)
                c.offset_nd.extend([s, e])



import numpy as np

from ..base import BaseNumericEncoder
from ...helper import batching, train_required


class StandarderEncoder(BaseNumericEncoder):
    batch_size = 2048

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mean = None
        self.scale = None

    def post_init(self):
        from sklearn.preprocessing import StandardScaler
        self.standarder = StandardScaler()

    @batching
    def train(self, vecs: np.ndarray, *args, **kwargs) -> None:
        self.standarder.partial_fit(vecs)

        self.mean = self.standarder.mean_.astype('float32')
        self.scale = self.standarder.scale_.astype('float32')

    @train_required
    @batching
    def encode(self, vecs: np.ndarray, *args, **kwargs) -> np.ndarray:
        return (vecs - self.mean) / self.scale



import os
from typing import Tuple, List, Union

import grpc
import zmq
from termcolor import colored

from ..helper import set_logger
from ..proto import gnes_pb2_grpc
from ..proto import send_message as _send_message, gnes_pb2, recv_message as _recv_message
from ..service.base import build_socket


class ResponseHandler:

    def __init__(self, h: 'ResponseHandler' = None):
        self.routes = {k: v for k, v in h.routes.items()} if h else {}
        self.logger = set_logger(self.__class__.__name__)
        self._context = None

    def register(self, resp_type: Union[List, Tuple, type]):

        def decorator(f):
            if isinstance(resp_type, list) or isinstance(resp_type, tuple):
                for t in resp_type:
                    self.routes[t] = f
            else:
                self.routes[resp_type] = f
            return f

        return decorator

    def call_routes(self, resp: 'gnes_pb2.Response'):

        def get_default_fn(r_type):
            self.logger.warning(
                'cant find handler for response type: %s, fall back to the default handler'
                % r_type)
            f = self.routes.get(r_type, self.routes[NotImplementedError])
            return f

        self.logger.info(
            'received a response for request %d' % resp.request_id)
        if resp.WhichOneof('body'):
            body = getattr(resp, resp.WhichOneof('body'))
            resp_type = type(body)

            if resp_type in self.routes:
                fn = self.routes.get(resp_type)
            else:
                fn = get_default_fn(type(resp))
            self.logger.info('handling response with %s' % fn.__name__)
            return fn(self._context, resp)
        else:
            self.logger.warning('the received message is not response')
            return None


class ZmqClient:

    def __init__(self, args):
        self.args = args
        self.logger = set_logger(self.__class__.__name__, self.args.verbose)
        self.ctx = zmq.Context()
        self.ctx.setsockopt(zmq.LINGER, 0)
        self.logger.info('current libzmq version is %s,  pyzmq version is %s' % (zmq.zmq_version(), zmq.__version__))

        self.receiver, recv_addr = build_socket(
            self.ctx, self.args.host_in, self.args.port_in,
            self.args.socket_in, self.args.identity)
        self.sender, send_addr = build_socket(self.ctx, self.args.host_out,
                                              self.args.port_out,
                                              self.args.socket_out,
                                              self.args.identity)
        self.logger.info(
            'input %s:%s\t output %s:%s' %
            (self.args.host_in, colored(self.args.port_in, 'yellow'),
             self.args.host_out, colored(self.args.port_out, 'yellow')))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.sender.close()
        self.receiver.close()
        self.ctx.term()

    def send_message(self, message: "gnes_pb2.Message", **kwargs):
        self.logger.debug('send message: %s' % message.envelope)
        _send_message(self.sender, message, **kwargs)

    def recv_message(self, **kwargs) -> gnes_pb2.Message:
        r = _recv_message(self.receiver, **kwargs)
        self.logger.debug('recv a message: %s' % r.envelope)
        return r


class GrpcClient:
    """
    A Base Unary gRPC client which the other client application can build from.

    """

    handler = ResponseHandler()

    def __init__(self, args):
        self.args = args
        if not args.proxy:
            os.unsetenv('http_proxy')
            os.unsetenv('https_proxy')
        self.logger = set_logger(self.__class__.__name__, self.args.verbose)
        self.logger.info('setting up grpc insecure channel...')
        # A gRPC channel provides a connection to a remote gRPC server.
        self._channel = grpc.insecure_channel(
            '%s:%d' % (self.args.grpc_host, self.args.grpc_port),
            options={
                'grpc.max_send_message_length': -1,
                'grpc.max_receive_message_length': -1,
            }.items(),
        )
        self.logger.info('waiting channel to be ready...')
        grpc.channel_ready_future(self._channel).result()

        # create new stub
        self.logger.info('create new stub...')
        self._stub = gnes_pb2_grpc.GnesRPCStub(self._channel)

        # attache response handler
        self.handler._context = self
        self.logger.critical('gnes client ready at %s:%d!' % (self.args.grpc_host, self.args.grpc_port))

    def call(self, request):
        resp = self._stub.call(request)
        self.handler.call_routes(resp)
        return resp

    def stream_call(self, request_iterator):
        response_stream = self._stub.StreamCall(request_iterator)
        for resp in response_stream:
            self.handler.call_routes(resp)

    @handler.register(NotImplementedError)
    def _handler_default(self, msg: 'gnes_pb2.Response'):
        raise NotImplementedError

    @handler.register(gnes_pb2.Response)
    def _handler_response_default(self, msg: 'gnes_pb2.Response'):
        pass

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def start(self):
        pass

    def close(self):
        self._channel.close()
        self._stub = None



# A key-value map for Class to the (module)file it located in
from ..base import register_all_class

_cls2file_map = {
    'FaissIndexer': 'chunk.faiss',
    'LVDBIndexer': 'doc.leveldb',
    'RocksDBIndexer': 'doc.rocksdb',
    'AsyncLVDBIndexer': 'doc.leveldb',
    'NumpyIndexer': 'chunk.numpy',
    'BIndexer': 'chunk.bindexer',
    'HBIndexer': 'chunk.hbindexer',
    'JointIndexer': 'base',
    'BaseIndexer': 'base',
    'BaseDocIndexer': 'base',
    'AnnoyIndexer': 'chunk.annoy',
    'DirectoryIndexer': 'doc.filesys',
    'DictIndexer': 'doc.dict',
    'DictKeyIndexer': 'chunk.helper',
    'ListKeyIndexer': 'chunk.helper',
    'ListNumpyKeyIndexer': 'chunk.helper',
    'NumpyKeyIndexer': 'chunk.helper',
}

register_all_class(_cls2file_map, 'indexer')



import unittest

from gnes.uuid import BaseIDGenerator, SnowflakeIDGenerator

class TestUUID(unittest.TestCase):
    def test_base_uuid(self):
        uuid_generator = BaseIDGenerator()
        last = -1
        for _ in range(10000):
            nid = uuid_generator.next()
            self.assertGreater(nid, last)
            last = nid


    def test_snoflake(self):
        uuid_generator = SnowflakeIDGenerator()
        last = -1
        for _ in range(10000):
            nid = uuid_generator.next()
            self.assertGreater(nid, last)
            last = nid



import os
import threading
from concurrent.futures import ThreadPoolExecutor

import grpc
from google.protobuf.json_format import MessageToJson

from ..client.base import ZmqClient
from ..helper import set_logger, make_route_table
from ..proto import gnes_pb2_grpc, gnes_pb2, router2str, add_route, add_version


class FrontendService:

    def __init__(self, args):
        if not args.proxy:
            os.unsetenv('http_proxy')
            os.unsetenv('https_proxy')
        self.logger = set_logger(self.__class__.__name__, args.verbose)
        self.server = grpc.server(
            ThreadPoolExecutor(max_workers=args.max_concurrency),
            options=[('grpc.max_send_message_length', args.max_message_size),
                     ('grpc.max_receive_message_length', args.max_message_size)])
        self.logger.info('start a frontend with %d workers' % args.max_concurrency)
        gnes_pb2_grpc.add_GnesRPCServicer_to_server(self._Servicer(args), self.server)

        self.bind_address = '{0}:{1}'.format(args.grpc_host, args.grpc_port)
        self.server.add_insecure_port(self.bind_address)
        self._stop_event = threading.Event()

    def __enter__(self):
        self.server.start()
        self.logger.critical('listening at: %s' % self.bind_address)
        self._stop_event.clear()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.server.stop(None)
        self.stop()

    def stop(self):
        self._stop_event.set()

    def join(self):
        self._stop_event.wait()

    class _Servicer(gnes_pb2_grpc.GnesRPCServicer):

        def __init__(self, args):
            self.args = args
            self.logger = set_logger(FrontendService.__name__, args.verbose)
            self.zmq_context = self.ZmqContext(args)
            self.request_id_cnt = 0
            self.send_recv_kwargs = dict(
                check_version=self.args.check_version,
                timeout=self.args.timeout,
                squeeze_pb=self.args.squeeze_pb)
            self.pending_request = 0

        def add_envelope(self, body: 'gnes_pb2.Request', zmq_client: 'ZmqClient'):
            msg = gnes_pb2.Message()
            msg.envelope.client_id = zmq_client.args.identity
            if body.request_id is not None:
                msg.envelope.request_id = body.request_id
            else:
                msg.envelope.request_id = self.request_id_cnt
                self.request_id_cnt += 1
                self.logger.warning('request_id is missing, filled it with an internal counter!')
            msg.envelope.part_id = 1
            msg.envelope.num_part.append(1)
            msg.envelope.timeout = 5000
            add_version(msg.envelope)
            add_route(msg.envelope, FrontendService.__name__, self.args.identity)
            msg.request.CopyFrom(body)
            return msg

        def remove_envelope(self, m: 'gnes_pb2.Message'):
            resp = m.response
            resp.request_id = m.envelope.request_id
            m.envelope.routes[0].end_time.GetCurrentTime()
            if self.args.route_table:
                self.logger.info('route: %s' % router2str(m))
                self.logger.info('route table: \n%s' % make_route_table(m.envelope.routes, include_frontend=True))
            if self.args.dump_route:
                self.args.dump_route.write(MessageToJson(m.envelope, indent=0).replace('\n', '') + '\n')
                self.args.dump_route.flush()
            return resp

        def Call(self, request, context):
            with self.zmq_context as zmq_client:
                zmq_client.send_message(self.add_envelope(request, zmq_client), **self.send_recv_kwargs)
                return self.remove_envelope(zmq_client.recv_message(**self.send_recv_kwargs))

        def Train(self, request, context):
            return self.Call(request, context)

        def Index(self, request, context):
            return self.Call(request, context)

        def Search(self, request, context):
            return self.Call(request, context)

        def StreamCall(self, request_iterator, context):
            self.pending_request = 0

            def get_response(num_recv, blocked=False):
                if blocked:
                    self.logger.info('waiting for %d responses ...' % (num_recv))
                for _ in range(num_recv):
                    if blocked or zmq_client.receiver.poll(1):
                        msg = zmq_client.recv_message(**self.send_recv_kwargs)
                        self.pending_request -= 1
                        yield self.remove_envelope(msg)

                while zmq_client.receiver.poll(1):
                    msg = zmq_client.recv_message(**self.send_recv_kwargs)
                    self.pending_request -= 1
                    yield self.remove_envelope(msg)

            with self.zmq_context as zmq_client:

                for request in request_iterator:
                    self.logger.info('receive request: %s' % request.request_id)
                    num_recv = max(self.pending_request - self.args.max_pending_request, 1)
                    yield from get_response(num_recv, num_recv > 1)
                    self.logger.info('send new request into %d appending tasks' % (self.pending_request))
                    zmq_client.send_message(self.add_envelope(request, zmq_client), **self.send_recv_kwargs)
                    self.pending_request += 1

                self.logger.info('all requests are sent, waiting for the responses...')
                yield from get_response(self.pending_request, blocked=True)

        class ZmqContext:
            """The zmq context class."""

            def __init__(self, args):
                self.args = args
                self.tlocal = threading.local()
                self.tlocal.client = None

            def __enter__(self):
                """Enter the context."""
                client = ZmqClient(self.args)
                self.tlocal.client = client
                return client

            def __exit__(self, exc_type, exc_value, exc_traceback):
                """Exit the context."""
                self.tlocal.client.close()
                self.tlocal.client = None



import sys
from termcolor import colored

from . import api
from .parser import get_main_parser

__all__ = ['main']


def get_run_args(parser_fn=get_main_parser, printed=True):
    parser = parser_fn()
    if len(sys.argv) > 1:
        args = parser.parse_args()
        if printed:
            param_str = '\n'.join(['%20s = %s' % (colored(k, 'yellow'), v) for k, v in sorted(vars(args).items())])
            print('usage: %s\n%s\n%s\n' % (' '.join(sys.argv), '_' * 50, param_str))
        return args
    else:
        parser.print_help()
        exit()


def main():
    args = get_run_args()
    getattr(api, args.cli, no_cli_error)(args)


def no_cli_error(*args, **kwargs):
    get_main_parser().print_help()
    exit()



import json
import unittest
from pprint import pprint

from gnes.proto import gnes_pb2
from gnes.score_fn.base import get_unary_score, CombinedScoreFn, ModifierScoreFn
from gnes.score_fn.chunk import WeightedChunkScoreFn, WeightedChunkOffsetScoreFn, CoordChunkScoreFn, TFIDFChunkScoreFn, BM25ChunkScoreFn
from gnes.score_fn.doc import CoordDocScoreFn
from gnes.score_fn.normalize import Normalizer1, Normalizer2, Normalizer3, Normalizer4

from gnes.cli.parser import set_router_parser, _set_client_parser
from gnes.service.base import SocketType
from gnes.client.base import ZmqClient
from gnes.service.router import RouterService


class TestScoreFn(unittest.TestCase):
    def test_basic(self):
        a = get_unary_score(0.5)
        b = get_unary_score(0.7)
        print(a)
        print(b.explained)

    def test_op(self):
        a = get_unary_score(0.5)
        b = get_unary_score(0.7)
        sum_op = CombinedScoreFn(score_mode='sum')
        c = sum_op(a, b)
        self.assertAlmostEqual(c.value, 1.2)

        sq_op = ModifierScoreFn(modifier='square')
        c = sum_op(a, sq_op(b))
        self.assertAlmostEqual(c.value, 0.99)
        print(c)

    def test_combine_score_fn(self):
        from gnes.indexer.chunk.helper import ListKeyIndexer
        from gnes.indexer.chunk.numpy import NumpyIndexer
        from gnes.proto import array2blob
        import numpy as np

        q_chunk = gnes_pb2.Chunk()
        q_chunk.doc_id = 2
        q_chunk.weight = 0.3
        q_chunk.offset = 0
        q_chunk.embedding.CopyFrom(array2blob(np.array([3, 3, 3])))

        for _fn in [WeightedChunkOffsetScoreFn, CoordChunkScoreFn, TFIDFChunkScoreFn, BM25ChunkScoreFn]:
            indexer = NumpyIndexer(helper_indexer=ListKeyIndexer(), score_fn=_fn())
            indexer.add(keys=[(0, 1), (1, 2)], vectors=np.array([[1, 1, 1], [2, 2, 2]]), weights=[0.5, 0.8])
            queried_result = indexer.query_and_score(q_chunks=[q_chunk], top_k=2)

    def test_doc_combine_score_fn(self):
        from gnes.indexer.doc.dict import DictIndexer

        document_list = []
        document_id_list = []

        for j in range(1, 4):
            d = gnes_pb2.Document()
            for i in range(1, 4):
                c = d.chunks.add()
                c.doc_id = j
                c.offset = i
                c.weight = 1 / 3
            document_id_list.append(j)
            document_list.append(d)

        self.chunk_router_yaml = 'Chunk2DocTopkReducer'

        args = set_router_parser().parse_args([
            '--yaml_path', self.chunk_router_yaml,
            '--socket_out', str(SocketType.PUB_BIND)
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.SUB_CONNECT)
        ])
        with RouterService(args), ZmqClient(c_args) as c1:
            msg = gnes_pb2.Message()
            s = msg.response.search.topk_results.add()
            s.score.value = 0.1
            s.score.explained = '"1-c1"'
            s.chunk.doc_id = 1

            s = msg.response.search.topk_results.add()
            s.score.value = 0.2
            s.score.explained = '"1-c2"'
            s.chunk.doc_id = 2

            s = msg.response.search.topk_results.add()

            s.score.value = 0.3
            s.score.explained = '"1-c3"'
            s.chunk.doc_id = 1

            msg.envelope.num_part.extend([1, 2])
            c1.send_message(msg)

            msg.response.search.ClearField('topk_results')

            s = msg.response.search.topk_results.add()
            s.score.value = 0.2
            s.score.explained = '"2-c1"'
            s.chunk.doc_id = 1

            s = msg.response.search.topk_results.add()
            s.score.value = 0.2
            s.score.explained = '"2-c2"'
            s.chunk.doc_id = 2

            s = msg.response.search.topk_results.add()
            s.score.value = 0.3
            s.score.explained = '"2-c3"'
            s.chunk.doc_id = 3
            c1.send_message(msg)
            r = c1.recv_message()
            doc_indexer = DictIndexer(score_fn=CoordDocScoreFn())
            doc_indexer.add(keys=document_id_list, docs=document_list)

            queried_result = doc_indexer.query_and_score(docs=r.response.search.topk_results, top_k=2)

    def test_normalizer(self):
        a = get_unary_score(0.5)
        norm_op = Normalizer1()
        b = norm_op(a)
        pprint(json.loads(b.explained))

        a = get_unary_score(0.5)
        norm_op = Normalizer2(2)
        b = norm_op(a)
        pprint(json.loads(b.explained))
        self.assertAlmostEqual(b.value, 0.8)

        a = get_unary_score(0.5)
        norm_op = Normalizer3(2)
        b = norm_op(a)
        pprint(json.loads(b.explained))
        self.assertAlmostEqual(b.value, 0.7387961283389092)

        a = get_unary_score(0.5)
        norm_op = Normalizer4(2)
        b = norm_op(a)
        pprint(json.loads(b.explained))
        self.assertEqual(b.value, 0.75)

        norm_op = ModifierScoreFn('none')
        b = norm_op(a)
        pprint(json.loads(b.explained))
        self.assertEqual(b.value, 0.5)

        q_chunk = gnes_pb2.Chunk()
        q_chunk.weight = 0.5
        q_chunk.offset = 1
        d_chunk = gnes_pb2.Chunk()
        d_chunk.weight = 0.7
        d_chunk.offset = 2
        rel_score = get_unary_score(2)
        _op = WeightedChunkScoreFn()
        c = _op(rel_score, q_chunk, d_chunk)
        pprint(json.loads(c.explained))
        self.assertAlmostEqual(c.value, 0.7)



import os
from typing import List

from ..base import BaseDocIndexer as BDI
from ...proto import gnes_pb2


class DirectoryIndexer(BDI):

    def __init__(self, data_path: str,
                 keep_na_doc: bool = True,
                 file_suffix: str = 'gif',
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_path = data_path
        self.file_suffix = file_suffix
        self.keep_na_doc = keep_na_doc
        self._NOT_FOUND = None

    @BDI.update_counter
    def add(self, keys: List[int], docs: List['gnes_pb2.Document'], *args, **kwargs):
        """
        write GIFs of each document into disk
        folder structure: /data_path/doc_id/0.gif, 1.gif...

        :param keys: list of doc id
        :param docs: list of docs
        """
        for k, d in zip(keys, docs):
            dirs = os.path.join(self.data_path, str(k))
            if not os.path.exists(dirs):
                os.makedirs(dirs)
            # keep doc meta in .meta file
            with open(os.path.join(dirs, '.meta'), 'wb') as f:
                f.write(d.meta_info or b'')

            for i, chunk in enumerate(d.chunks):
                with open(os.path.join(dirs, '%d.%s' % (i, self.file_suffix)), 'wb') as f:
                    f.write(chunk.raw)

    def query(self, keys: List[int], *args, **kwargs) -> List['gnes_pb2.Document']:
        """
        Find the doc according to the keys

        :param keys: list of doc id
        :return: list of documents whose chunks field contain all the GIFs of this doc(one GIF per chunk)
        """
        res = []
        for k in keys:
            doc = gnes_pb2.Document()
            target_dirs = os.path.join(self.data_path, str(k))

            if not os.path.exists(target_dirs):
                if self.keep_na_doc:
                    res.append(self._NOT_FOUND)
            else:
                with open(os.path.join(target_dirs, '.meta'), 'rb') as f:
                    doc.meta_info = f.read()
                for raw_file in os.listdir(target_dirs):
                    if not os.path.isdir(raw_file):
                        c = doc.chunks.add()
                        c.doc_id = k
                        with open(os.path.join(target_dirs, raw_file), 'rb') as raw:
                            c.raw = raw.read()
                res.append(doc)
        return res



import numpy as np

from .base import BaseService as BS, MessageHandler, ServiceError
from ..proto import gnes_pb2, blob2array


class IndexerService(BS):
    handler = MessageHandler(BS.handler)

    def post_init(self):
        from ..indexer.base import BaseIndexer
        # print('id: %s, before: %r' % (threading.get_ident(), self._model))
        self._model = self.load_model(BaseIndexer)
        # self._tmp_a = threading.get_ident()
        # print('id: %s, after: %r, self._tmp_a: %r' % (threading.get_ident(), self._model, self._tmp_a))

    @handler.register(gnes_pb2.Request.IndexRequest)
    def _handler_index(self, msg: 'gnes_pb2.Message'):
        # print('tid: %s, model: %r, self._tmp_a: %r' % (threading.get_ident(), self._model, self._tmp_a))
        # if self._tmp_a != threading.get_ident():
        #     print('!!! tid: %s, tmp_a: %r %r' % (threading.get_ident(), self._tmp_a, self._handler_index))
        from ..indexer.base import BaseChunkIndexer, BaseDocIndexer
        if isinstance(self._model, BaseChunkIndexer):
            is_changed = self._handler_chunk_index(msg)
        elif isinstance(self._model, BaseDocIndexer):
            is_changed = self._handler_doc_index(msg)
        else:
            raise ServiceError(
                'unsupported indexer, dont know how to use %s to handle this message' % self._model.__bases__)

        if self.args.as_response:
            msg.response.index.status = gnes_pb2.Response.SUCCESS

        if is_changed:
            self.is_model_changed.set()

    def _handler_chunk_index(self, msg: 'gnes_pb2.Message') -> bool:
        embed_info = []

        for d in msg.request.index.docs:
            if not d.chunks:
                self.logger.warning('document (doc_id=%s) contains no chunks!' % d.doc_id)
                continue

            embed_info += [(blob2array(c.embedding), d.doc_id, c.offset, c.weight) for c in d.chunks if
                           c.embedding.data]

        if embed_info:
            vecs, doc_ids, offsets, weights = zip(*embed_info)
            self._model.add(list(zip(doc_ids, offsets)), np.stack(vecs), weights)
            return True
        else:
            self.logger.warning('chunks contain no embedded vectors, the indexer will do nothing')
            return False

    def _handler_doc_index(self, msg: 'gnes_pb2.Message') -> bool:
        if msg.request.index.docs:
            self._model.add([d.doc_id for d in msg.request.index.docs],
                            [d for d in msg.request.index.docs],
                            [d.weight for d in msg.request.index.docs])
            return True
        else:
            return False

    def _put_result_into_message(self, results, msg: 'gnes_pb2.Message'):
        msg.response.search.ClearField('topk_results')
        msg.response.search.topk_results.extend(results)
        msg.response.search.top_k = len(results)
        msg.response.search.is_big_score_similar = self._model.is_big_score_similar

    @handler.register(gnes_pb2.Request.QueryRequest)
    def _handler_chunk_search(self, msg: 'gnes_pb2.Message'):
        from ..indexer.base import BaseChunkIndexer
        if not isinstance(self._model, BaseChunkIndexer):
            raise ServiceError(
                'unsupported indexer, dont know how to use %s to handle this message' % self._model.__bases__)

        results = []
        if not msg.request.search.query.chunks:
            self.logger.warning('query contains no chunks!')
        else:
            results = self._model.query_and_score(msg.request.search.query.chunks, top_k=msg.request.search.top_k)

        self._put_result_into_message(results, msg)

    @handler.register(gnes_pb2.Response.QueryResponse)
    def _handler_doc_search(self, msg: 'gnes_pb2.Message'):
        from ..indexer.base import BaseDocIndexer
        if not isinstance(self._model, BaseDocIndexer):
            raise ServiceError(
                'unsupported indexer, dont know how to use %s to handle this message' % self._model.__bases__)

        # check if chunk_indexer and doc_indexer has the same sorting order
        if msg.response.search.is_big_score_similar is not None and \
                msg.response.search.is_big_score_similar != self._model.is_big_score_similar:
            raise ServiceError(
                'is_big_score_similar is inconsistent. last topk-list: is_big_score_similar=%s, but '
                'this indexer: is_big_score_similar=%s' % (
                    msg.response.search.is_big_score_similar, self._model.is_big_score_similar))

        # assume the doc search will change the whatever sort order the message has
        msg.response.search.is_sorted = False
        results = self._model.query_and_score(msg.response.search.topk_results)
        self._put_result_into_message(results, msg)



from typing import List

import numpy as np

from ..base import BaseTextEncoder
from ...helper import batching, as_numpy_array


class CharEmbeddingEncoder(BaseTextEncoder):
    """A random character embedding model. Only useful for testing"""
    is_trained = True

    def __init__(self, dim: int = 128, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dim = dim
        self.offset = 32
        self.unknown_idx = 96
        # in total 96 printable chars and 2 special chars = 98
        self._char_embedding = np.random.random([97, dim])

    @batching
    @as_numpy_array
    def encode(self, text: List[str], *args, **kwargs) -> List[np.ndarray]:
        # tokenize text
        sent_embed = []
        for sent in text:
            ids = [ord(c) - 32 if 32 <= ord(c) <= 127 else self.unknown_idx for c in sent]
            sent_embed.append(np.mean(self._char_embedding[ids], axis=0))
        return sent_embed



import os
import unittest

import numpy as np

from gnes.cli.parser import set_encoder_parser, _set_client_parser
from gnes.client.base import ZmqClient
from gnes.proto import gnes_pb2, array2blob
from gnes.service.base import ServiceManager
from gnes.service.encoder import EncoderService
from gnes.encoder.base import BaseEncoder


class TestEncoder(BaseEncoder):

    def encode(self, x):
        return np.array(x)


class TestEncoderService(unittest.TestCase):

    def setUp(self):
        self.test_numeric = np.random.randint(0, 255, (1000, 1024)).astype('float32')

    def test_empty_service(self):
        args = set_encoder_parser().parse_args(['--yaml_path', '!TestEncoder {gnes_config: {name: EncoderService, is_trained: true}}'])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)])

        with ServiceManager(EncoderService, args), ZmqClient(c_args) as client:
            msg = gnes_pb2.Message()
            d = msg.request.index.docs.add()
            d.doc_type = gnes_pb2.Document.IMAGE

            c = d.chunks.add()
            c.blob.CopyFrom(array2blob(self.test_numeric))

            client.send_message(msg)
            r = client.recv_message()
            self.assertEqual(len(r.request.index.docs), 1)
            self.assertEqual(r.response.index.status, gnes_pb2.Response.SUCCESS)

    def tearDown(self):
        if os.path.exists('EncoderService.bin'):
            os.remove('EncoderService.bin')



from .base import YamlComposer, parse_http_data
from ..cli.parser import set_composer_parser
from ..helper import set_logger


class YamlComposerFlask:
    def __init__(self, args):
        self.args = args
        self.logger = set_logger(self.__class__.__name__, self.args.verbose)

    def _create_flask_app(self):
        try:
            from flask import Flask, request
        except ImportError:
            raise ImportError('Flask or its dependencies are not fully installed, '
                              'they are required for serving HTTP requests.'
                              'Please use "pip install gnes[flask]" to install it.')

        app = Flask(__name__)
        args = set_composer_parser().parse_args([])
        default_html = YamlComposer(args).build_all()['html']

        @app.errorhandler(500)
        def exception_handler(error):
            self.logger.error('unhandled error, i better quit and restart! %s' % error)
            return '<h1>500 Internal Error</h1> ' \
                   'While we are fixing the issue, do you know you can deploy GNES board locally on your machine? ' \
                   'Simply run <pre>docker run -d -p 0.0.0.0:80:8080/tcp gnes/gnes compose --flask</pre>', 500

        @app.route('/', methods=['GET'])
        def _get_homepage():
            return default_html

        @app.route('/generate', methods=['POST'])
        def _regenerate():
            data = request.form if request.form else request.json
            return parse_http_data(data, args)

        return app

    def run(self):
        app = self._create_flask_app()
        app.run(port=self.args.http_port, threaded=True, host='0.0.0.0')



from typing import List, Any, Tuple, Union

import numpy as np

from ..base import TrainableBase, CompositionalTrainableBase


class BaseEncoder(TrainableBase):

    def encode(self, data: Any, *args, **kwargs) -> Any:
        pass

    def _copy_from(self, x: 'BaseEncoder') -> None:
        pass


class BaseImageEncoder(BaseEncoder):

    def encode(self, img: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        pass


class BaseVideoEncoder(BaseEncoder):

    def encode(self, data: List['np.ndarray'], *args, **kwargs) -> Union[np.ndarray, List['np.ndarray']]:
        pass


class BaseTextEncoder(BaseEncoder):

    def encode(self, text: List[str], *args, **kwargs) -> Union[Tuple, np.ndarray]:
        pass


class BaseNumericEncoder(BaseEncoder):
    """Note that all NumericEncoder can not be used as the first encoder of the pipeline"""

    def encode(self, data: np.ndarray, *args, **kwargs) -> np.ndarray:
        pass


class BaseAudioEncoder(BaseEncoder):

    def encode(self, data: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        pass


class BaseBinaryEncoder(BaseEncoder):

    def encode(self, data: np.ndarray, *args, **kwargs) -> bytes:
        if data.dtype != np.uint8:
            raise ValueError('data must be np.uint8 but received %s' % data.dtype)
        return data.tobytes()


class PipelineEncoder(CompositionalTrainableBase):
    def encode(self, data: Any, *args, **kwargs) -> Any:
        if not self.components:
            raise NotImplementedError
        for be in self.components:
            data = be.encode(data, *args, **kwargs)
        return data

    def train(self, data, *args, **kwargs):
        if not self.components:
            raise NotImplementedError
        for idx, be in enumerate(self.components):
            if not be.is_trained:
                be.train(data, *args, **kwargs)

            if idx + 1 < len(self.components):
                data = be.encode(data, *args, **kwargs)



from collections import defaultdict
from functools import wraps
from typing import List, Any, Union, Callable, Tuple

import numpy as np

from ..base import TrainableBase, CompositionalTrainableBase
from ..proto import gnes_pb2, blob2array
from ..score_fn.base import get_unary_score, ModifierScoreFn


class BaseIndexer(TrainableBase):
    def __init__(self,
                 normalize_fn: 'BaseScoreFn' = None,
                 score_fn: 'BaseScoreFn' = None,
                 is_big_score_similar: bool = False,
                 *args, **kwargs):
        """
        Base indexer, a valid indexer must implement :py:meth:`add` and :py:meth:`query` methods

        :type score_fn: advanced score function
        :type normalize_fn: normalizing score function
        :type is_big_score_similar: when set to true, then larger score means more similar
        """
        super().__init__(*args, **kwargs)
        self.normalize_fn = normalize_fn if normalize_fn else ModifierScoreFn()
        self.score_fn = score_fn if score_fn else ModifierScoreFn()
        self.normalize_fn._context = self
        self.score_fn._context = self
        self.is_big_score_similar = is_big_score_similar
        self._num_docs = 0
        self._num_chunks = 0
        self._num_chunks_in_doc = defaultdict(int)

    def add(self, keys: Any, docs: Any, weights: List[float], *args, **kwargs):
        pass

    def query(self, keys: Any, *args, **kwargs) -> List[Any]:
        pass

    def query_and_score(self, q_chunks: List[Union['gnes_pb2.Chunk', 'gnes_pb2.Document']], top_k: int) -> List[
        'gnes_pb2.Response.QueryResponse.ScoredResult']:
        raise NotImplementedError

    @property
    def num_docs(self):
        return self._num_docs

    @property
    def num_chunks(self):
        return self._num_chunks


class BaseChunkIndexer(BaseIndexer):
    """Storing chunks and their vector representations """

    def __init__(self, helper_indexer: 'BaseChunkIndexerHelper' = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper_indexer = helper_indexer

    def add(self, keys: List[Tuple[int, int]], vectors: np.ndarray, weights: List[float], *args, **kwargs):
        """
        adding new chunks and their vector representations

        :param keys: list of (doc_id, offset) tuple
        :param vectors: vector representations
        :param weights: weight of the chunks
        """
        pass

    def query(self, keys: np.ndarray, top_k: int, *args, **kwargs) -> List[List[Tuple]]:
        pass

    def query_and_score(self, q_chunks: List['gnes_pb2.Chunk'], top_k: int, *args, **kwargs) -> List[
        'gnes_pb2.Response.QueryResponse.ScoredResult']:
        vecs = [blob2array(c.embedding) for c in q_chunks]
        queried_results = self.query(np.stack(vecs), top_k=top_k)
        results = []
        for q_chunk, topk_chunks in zip(q_chunks, queried_results):
            for _doc_id, _offset, _weight, _relevance in topk_chunks:
                r = gnes_pb2.Response.QueryResponse.ScoredResult()
                r.chunk.doc_id = _doc_id
                r.chunk.offset = _offset
                r.chunk.weight = _weight
                _score = get_unary_score(value=_relevance,
                                         name=self.__class__.__name__,
                                         operands=[
                                             dict(name='doc_chunk',
                                                  doc_id=_doc_id,
                                                  offset=_offset),
                                             dict(name='query_chunk',
                                                  offset=q_chunk.offset)])
                _score = self.normalize_fn(_score)
                _score = self.score_fn(_score, q_chunk, r.chunk, queried_results)
                r.score.CopyFrom(_score)
                results.append(r)
        return results

    @staticmethod
    def update_counter(func):
        @wraps(func)
        def arg_wrapper(self, keys: List[Tuple[int, int]], *args, **kwargs):
            doc_ids, _ = zip(*keys)
            self._num_docs += len(set(doc_ids))
            self._num_chunks += len(keys)
            for doc_id in doc_ids:
                self._num_chunks_in_doc[doc_id] += 1
            return func(self, keys, *args, **kwargs)

        return arg_wrapper

    @staticmethod
    def update_helper_indexer(func):
        @wraps(func)
        def arg_wrapper(self, keys: List[Tuple[int, int]], vectors: np.ndarray, weights: List[float], *args, **kwargs):
            r = func(self, keys, vectors, weights, *args, **kwargs)
            if self.helper_indexer:
                self.helper_indexer.add(keys, weights, *args, **kwargs)
            return r

        return arg_wrapper

    @property
    def num_docs(self):
        if self.helper_indexer:
            return self.helper_indexer._num_docs
        else:
            return self._num_docs

    @property
    def num_chunks(self):
        if self.helper_indexer:
            return self.helper_indexer._num_chunks
        else:
            return self._num_chunks

    def num_chunks_in_doc(self, doc_id: int):
        if self.helper_indexer:
            return self.helper_indexer._num_chunks_in_doc[doc_id]
        else:
            self.logger.warning('enable helper_indexer to track num_chunks_in_doc')


class BaseDocIndexer(BaseIndexer):
    """Storing documents and contents """

    def add(self, keys: List[int], docs: List['gnes_pb2.Document'], *args, **kwargs):
        """
        adding new docs and their protobuf representation

        :param keys: list of doc_id
        :param docs: list of protobuf Document objects
        """
        pass

    def query(self, keys: List[int], *args, **kwargs) -> List['gnes_pb2.Document']:
        pass

    def query_and_score(self, docs: List['gnes_pb2.Response.QueryResponse.ScoredResult'], *args, **kwargs) -> List[
        'gnes_pb2.Response.QueryResponse.ScoredResult']:
        keys = [r.doc.doc_id for r in docs]
        results = []
        queried_results = self.query(keys, *args, **kwargs)
        for d, r in zip(queried_results, docs):
            if d:
                r.doc.CopyFrom(d)
                _score = self.normalize_fn(r.score)
                _score = self.score_fn(_score, d)
                r.score.CopyFrom(_score)
            results.append(r)
        return results

    @staticmethod
    def update_counter(func):
        @wraps(func)
        def arg_wrapper(self, keys: List[int], docs: List['gnes_pb2.Document'], *args, **kwargs):
            self._num_docs += len(keys)
            self._num_chunks += sum(len(d.chunks) for d in docs)
            return func(self, keys, docs, *args, **kwargs)

        return arg_wrapper


class BaseChunkIndexerHelper(BaseChunkIndexer):
    """A helper class for storing chunk info, doc mapping, weights.
    This is especially useful when ChunkIndexer can not store these information by itself
    """

    def add(self, keys: List[Tuple[int, int]], weights: List[float], *args, **kwargs) -> int:
        pass

    def query(self, keys: List[int], *args, **kwargs) -> List[Tuple[int, int, float]]:
        pass


class JointIndexer(CompositionalTrainableBase):

    @property
    def components(self):
        return self._component

    @components.setter
    def components(self, comps: Callable[[], Union[list, dict]]):
        if not callable(comps):
            raise TypeError('component must be a callable function that returns '
                            'a List[BaseIndexer]')
        if not getattr(self, 'init_from_yaml', False):
            self._component = comps()
        else:
            self.logger.info('component is omitted from construction, '
                             'as it is initialized from yaml config')

        self._binary_indexer = None
        self._doc_indexer = None
        for c in self.components:
            if isinstance(c, BaseChunkIndexer):
                self._binary_indexer = c
            elif isinstance(c, BaseDocIndexer):
                self._doc_indexer = c
        if not self._binary_indexer or not self._doc_indexer:
            raise ValueError('"JointIndexer" requires a valid pair of "BaseBinaryIndexer" and "BaseTextIndexer"')

    def add(self, keys: Any, docs: Any, *args,
            **kwargs) -> None:
        if isinstance(docs, np.ndarray):
            self._binary_indexer.add(keys, docs, *args, **kwargs)
        elif isinstance(docs, list):
            self._doc_indexer.add(keys, docs, *args, **kwargs)
        else:
            raise TypeError('can not find an indexer for doc type: %s' % type(docs))

    def query(self,
              keys: Any,
              top_k: int,
              *args,
              **kwargs) -> List[List[Tuple]]:
        topk_results = self._binary_indexer.query(keys, top_k, *args, **kwargs)
        doc_caches = dict()
        topk_results_with_docs = []
        for topk in topk_results:
            topk_wd = []
            for doc_id, offset, weight, score in topk:
                doc = doc_caches.get(doc_id, self._doc_indexer.query([doc_id])[0])
                doc_caches[doc_id] = doc
                topk_wd.append((doc_id, offset, weight, score, doc.chunks[offset]))
            topk_results_with_docs.append(topk_wd)
        return topk_results_with_docs



import os
from typing import Tuple

import numpy as np

from ..base import BaseNumericEncoder
from ...helper import as_numpy_array


class PoolingEncoder(BaseNumericEncoder):
    def __init__(self, pooling_strategy: str = 'REDUCE_MEAN',
                 backend: str = 'numpy',
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        valid_poolings = {'REDUCE_MEAN', 'REDUCE_MAX', 'REDUCE_MEAN_MAX'}
        valid_backends = {'tensorflow', 'numpy', 'pytorch', 'torch'}

        if pooling_strategy not in valid_poolings:
            raise ValueError('"pooling_strategy" must be one of %s' % valid_poolings)
        if backend not in valid_backends:
            raise ValueError('"backend" must be one of %s' % valid_backends)
        self.pooling_strategy = pooling_strategy
        self.backend = backend

    def post_init(self):
        if self.backend in {'pytorch', 'torch'}:
            import torch
            self.torch = torch
        elif self.backend == 'tensorflow':
            os.environ['CUDA_VISIBLE_DEVICES'] = '0' if self.on_gpu else '-1'
            import tensorflow as tf
            self._tf_graph = tf.Graph()
            config = tf.ConfigProto(device_count={'GPU': 1 if self.on_gpu else 0})
            config.gpu_options.allow_growth = True
            config.log_device_placement = False
            self._sess = tf.Session(graph=self._tf_graph, config=config)
            self.tf = tf

    def mul_mask(self, x, m):
        if self.backend in {'pytorch', 'torch'}:
            return self.torch.mul(x, m.unsqueeze(2))
        elif self.backend == 'tensorflow':
            with self._tf_graph.as_default():
                return x * self.tf.expand_dims(m, axis=-1)
        elif self.backend == 'numpy':
            return x * np.expand_dims(m, axis=-1)

    def minus_mask(self, x, m, offset: int = 1e30):
        if self.backend in {'pytorch', 'torch'}:
            return x - (1.0 - m).unsqueeze(2) * offset
        elif self.backend == 'tensorflow':
            with self._tf_graph.as_default():
                return x - self.tf.expand_dims(1.0 - m, axis=-1) * offset
        elif self.backend == 'numpy':
            return x - np.expand_dims(1.0 - m, axis=-1) * offset

    def masked_reduce_mean(self, x, m, jitter: float = 1e-10):
        if self.backend in {'pytorch', 'torch'}:
            return self.torch.div(self.torch.sum(self.mul_mask(x, m), dim=1),
                                  self.torch.sum(m.unsqueeze(2), dim=1) + jitter)
        elif self.backend == 'tensorflow':
            with self._tf_graph.as_default():
                return self.tf.reduce_sum(self.mul_mask(x, m), axis=1) / (
                        self.tf.reduce_sum(m, axis=1, keepdims=True) + jitter)
        elif self.backend == 'numpy':
            return np.sum(self.mul_mask(x, m), axis=1) / (np.sum(m, axis=1, keepdims=True) + jitter)

    def masked_reduce_max(self, x, m):
        if self.backend in {'pytorch', 'torch'}:
            return self.torch.max(self.minus_mask(x, m), 1)[0]
        elif self.backend == 'tensorflow':
            with self._tf_graph.as_default():
                return self.tf.reduce_max(self.minus_mask(x, m), axis=1)
        elif self.backend == 'numpy':
            return np.max(self.minus_mask(x, m), axis=1)

    @as_numpy_array
    def encode(self, data: Tuple, *args, **kwargs):
        seq_tensor, mask_tensor = data

        if self.pooling_strategy == 'REDUCE_MEAN':
            r = self.masked_reduce_mean(seq_tensor, mask_tensor)
        elif self.pooling_strategy == 'REDUCE_MAX':
            r = self.masked_reduce_max(seq_tensor, mask_tensor)
        elif self.pooling_strategy == 'REDUCE_MEAN_MAX':
            if self.backend in {'pytorch', 'torch'}:
                r = self.torch.cat((self.masked_reduce_mean(seq_tensor, mask_tensor),
                                    self.masked_reduce_max(seq_tensor, mask_tensor)), dim=1)
            elif self.backend == 'tensorflow':
                with self._tf_graph.as_default():
                    r = self.tf.concat([self.masked_reduce_mean(seq_tensor, mask_tensor),
                                        self.masked_reduce_max(seq_tensor, mask_tensor)], axis=1)
            elif self.backend == 'numpy':
                r = np.concatenate([self.masked_reduce_mean(seq_tensor, mask_tensor),
                                    self.masked_reduce_max(seq_tensor, mask_tensor)], axis=1)

        if self.backend == 'tensorflow':
            r = self._sess.run(r)
        return r



# A key-value map for Class to the (module)file it located in
from ..base import register_all_class

_cls2file_map = {
    'BertEncoder': 'text.bert',
    'BertEncoderWithServer': 'text.bert',
    'BertEncoderServer': 'text.bert',
    'FlairEncoder': 'text.flair',
    'PCALocalEncoder': 'numeric.pca',
    'PQEncoder': 'numeric.pq',
    'TFPQEncoder': 'numeric.tf_pq',
    'Word2VecEncoder': 'text.w2v',
    'BaseEncoder': 'base',
    'BaseBinaryEncoder': 'base',
    'BaseTextEncoder': 'base',
    'BaseVideoEncoder': 'base',
    'BaseNumericEncoder': 'base',
    'BaseAudioEncoder': 'base',
    'PipelineEncoder': 'base',
    'HashEncoder': 'numeric.hash',
    'TorchvisionEncoder': 'image.torchvision',
    'TFInceptionEncoder': 'image.inception',
    'CVAEEncoder': 'image.cvae',
    'IncepMixtureEncoder': 'video.incep_mixture',
    'VladEncoder': 'numeric.vlad',
    'MfccEncoder': 'audio.mfcc',
    'PoolingEncoder': 'numeric.pooling',
    'PyTorchTransformers': 'text.transformer',
    'VggishEncoder': 'audio.vggish',
    'YouTube8MFeatureExtractor': 'video.yt8m_feature_extractor',
    'YouTube8MEncoder': 'video.yt8m_model',
    'InceptionVideoEncoder': 'video.inception',
    'QuantizerEncoder': 'numeric.quantizer',
    'CharEmbeddingEncoder': 'text.char'
}

register_all_class(_cls2file_map, 'encoder')



import re

from .helper import _check_input, kwargs_to_cmd_args, run_command

VIDEO_DUR_PATTERN = re.compile(r".*Duration: (\d+):(\d+):(\d+)", re.DOTALL)
VIDEO_INFO_PATTERN = re.compile(
    r'.*Stream #0:(\d+)(?:\(\w+\))?: Video: (\w+).*, (\w+)[(,].* (\d+)x(\d+).* (\d+)(\.\d.)? fps',
    re.DOTALL)
AUDIO_INFO_PATTERN = re.compile(
    r'^\s+Stream #0:(?P<stream>\d+)(\((?P<lang>\w+)\))?: Audio: (?P<format>\w+).*?(?P<default>\(default\))?$',
    re.MULTILINE)
STREAM_SUBTITLE_PATTERN = re.compile(
    r'^\s+Stream #0:(?P<stream>\d+)(\((?P<lang>\w+)\))?: Subtitle:',
    re.MULTILINE)


def parse_media_details(infos):
    video_dur_match = VIDEO_DUR_PATTERN.match(infos)
    dur_hrs, dur_mins, dur_secs = video_dur_match.group(1, 2, 3)

    video_info_match = VIDEO_INFO_PATTERN.match(infos)
    codec, pix_fmt, res_width, res_height, fps = video_info_match.group(
        2, 3, 4, 5, 6)

    audio_tracks = list()
    for audio_match in AUDIO_INFO_PATTERN.finditer(infos):
        ainfo = audio_match.groupdict()
        if ainfo['lang'] is None:
            ainfo['lang'] = 'und'
        audio_tracks.append(ainfo)

    medio_info = {
        'vcodec': codec,
        'frame_width': int(res_width),
        'frame_height': int(res_height),
        'duration': (int(dur_hrs) * 3600 + int(dur_mins) * 60 + int(dur_secs)),
        'fps': int(fps),
        'pix_fmt': pix_fmt,
        'audio': audio_tracks,
    }
    return medio_info


def extract_frame_size(ffmpeg_parse_info: str):
    possible_patterns = [
        re.compile(r'Stream.*Video.*([0-9]{4,})x([0-9]{4,})'),
        re.compile(r'Stream.*Video.*([0-9]{4,})x([0-9]{3,})'),
        re.compile(r'Stream.*Video.*([0-9]{3,})x([0-9]{3,})')
    ]

    for pattern in possible_patterns:
        match = pattern.search(ffmpeg_parse_info)
        if match is not None:
            x, y = map(int, match.groups()[0:2])
            break

    if match is None:
        raise ValueError("could not get video frame size")

    return (x, y)


def compile_args(input_fn: str = 'pipe:',
                 output_fn: str = 'pipe:',
                 video_filters: str = [],
                 audio_filters: str = [],
                 input_options=dict(),
                 output_options=dict(),
                 overwrite_output: bool = True):
    """Wrapper for various `FFmpeg <https://www.ffmpeg.org/>`_ related applications (ffmpeg,
    ffprobe).
    """
    args = ['ffmpeg', '-threads', '1']

    input_args = []
    fmt = input_options.pop('format', None)
    if fmt:
        input_args += ['-f', fmt]

    start_time = input_options.pop('ss', None)
    duration = input_options.pop('t', None)

    input_args += kwargs_to_cmd_args(input_options)
    input_args += ['-i', input_fn]
    if start_time is not None:
        input_args += ['-ss', str(start_time)]
    if duration is not None:
        input_args += ['-t', str(duration)]

    vf_args = []
    if len(video_filters) > 0:
        vf_args = ['-vf', ','.join(video_filters)]

    af_args = []
    if len(audio_filters) > 0:
        af_args = ['-af', ','.join(audio_filters)]

    output_args = []

    fmt = output_options.pop('format', None)
    if fmt:
        output_args += ['-f', fmt]
    video_bitrate = output_options.pop('video_bitrate', None)
    if video_bitrate:
        output_args += ['-b:v', str(video_bitrate)]
    audio_bitrate = output_options.pop('audio_bitrate', None)
    if audio_bitrate:
        output_args += ['-b:a', str(audio_bitrate)]
    output_args += kwargs_to_cmd_args(output_options)

    output_args += [output_fn]

    args += input_args + vf_args + af_args + output_args

    if overwrite_output:
        args += ['-y']

    return args


def probe(input_fn: str):
    command = [
        'ffprobe', '-v', 'fatal', '-show_entries',
        'stream=width,height,r_frame_rate,duration', '-of',
        'default=noprint_wrappers=1:nokey=1', input_fn, '-sexagesimal'
    ]
    out, err = run_command(command, pipe_stdout=True, pipe_stderr=True)

    out = out.decode().split('\n')
    return {
        'file': input_fn,
        'width': int(out[0]),
        'height': int(out[1]),
        'fps': float(out[2].split('/')[0]) / float(out[2].split('/')[1]),
        'duration': out[3]
    }


def get_media_meta(input_fn: str = 'pipe:',
                   input_data: bytes = None,
                   input_options=dict()):
    _check_input(input_fn, input_data)
    cmd_args = ['ffmpeg']

    fmt = input_options.pop('format', None)
    if fmt:
        cmd_args += ['-f', fmt]
    cmd_args += ['-i', input_fn]

    cmd_args += ['-f', 'ffmetadata', 'pipe:']
    out, err = run_command(
        cmd_args, input=input_data, pipe_stdout=True, pipe_stderr=True)
    return parse_media_details(err.decode())



from typing import List, Tuple

import torch

from ..base import BaseTextEncoder
from ...helper import batching


class PyTorchTransformers(BaseTextEncoder):
    is_trained = True

    def __init__(self,
                 model_name: str = 'bert-base-uncased',
                 pooling_layer: int = 0,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_name = model_name
        self.pooling_layer = pooling_layer

    def post_init(self):
        from pytorch_transformers import BertModel, BertTokenizer, \
            OpenAIGPTModel, OpenAIGPTTokenizer, GPT2Model, GPT2Tokenizer, \
            TransfoXLModel, TransfoXLTokenizer, XLNetModel, XLNetTokenizer, \
            XLMModel, XLMTokenizer, RobertaModel, RobertaTokenizer
        # select the model, tokenizer & weight accordingly
        model_class, tokenizer_class, pretrained_weights = \
            {k[-1]: k for k in
             [(BertModel, BertTokenizer, 'bert-base-uncased'),
              (OpenAIGPTModel, OpenAIGPTTokenizer, 'openai-gpt'),
              (GPT2Model, GPT2Tokenizer, 'gpt2'),
              (TransfoXLModel, TransfoXLTokenizer, 'transfo-xl-wt103'),
              (XLNetModel, XLNetTokenizer, 'xlnet-base-cased'),
              (XLMModel, XLMTokenizer, 'xlm-mlm-enfr-1024'),
              (RobertaModel, RobertaTokenizer, 'roberta-base')]}[self.model_name]

        def load_model_tokenizer(x):
            return model_class.from_pretrained(x).eval(), tokenizer_class.from_pretrained(x)

        try:
            self.model, self.tokenizer = load_model_tokenizer(self.work_dir)
        except Exception:
            self.logger.warning('cannot deserialize model/tokenizer from %s, will download from web' % self.work_dir)
            self.model, self.tokenizer = load_model_tokenizer(pretrained_weights)

    @batching
    def encode(self, text: List[str], *args, **kwargs) -> Tuple:
        # encoding and padding
        ids = [self.tokenizer.encode(t) for t in text]
        max_len = max(len(t) for t in ids)
        ids = [t + [0] * (max_len - len(t)) for t in ids]
        m_ids = [[1] * len(t) + [0] * (max_len - len(t)) for t in ids]
        seq_ids = torch.tensor(ids)
        mask_ids = torch.tensor(m_ids, dtype=torch.float32)

        if self.on_gpu:
            seq_ids = seq_ids.cuda()

        with torch.no_grad():
            last_hidden_states = self.model(seq_ids)[self.pooling_layer]

        return last_hidden_states.cpu(), mask_ids

    def __getstate__(self):
        self.model.save_pretrained(self.work_dir)
        self.tokenizer.save_pretrained(self.work_dir)
        return super().__getstate__()



import io
import re
from typing import List

import numpy as np
import soundfile as sf

from .ffmpeg import compile_args
from .helper import _check_input, run_command

DEFAULT_SILENCE_DURATION = 0.3
DEFAULT_SILENCE_THRESHOLD = -60


def capture_audio(input_fn: str = 'pipe:',
                  input_data: bytes = None,
                  bits_per_raw_sample: int = 16,
                  sample_rate: int = 16000,
                  start_time: float = None,
                  end_time: float = None,
                  **kwargs) -> List['np.ndarray']:
    _check_input(input_fn, input_data)

    input_kwargs = {}
    if start_time is not None:
        input_kwargs['ss'] = str(start_time)
    else:
        start_time = 0.
    if end_time is not None:
        input_kwargs['t'] = str(end_time - start_time)

    output_kwargs = {
        'format': 'wav',
        'bits_per_raw_sample': bits_per_raw_sample,
        'ac': 1,
        'ar': sample_rate
    }

    cmd_args = compile_args(
        input_fn=input_fn,
        input_options=input_kwargs,
        output_options=output_kwargs,
        overwrite_output=True)

    stdout, _ = run_command(
        cmd_args, input=input_data, pipe_stdout=True, pipe_stderr=True)

    audio_stream = io.BytesIO(stdout)
    audio_data, sample_rate = sf.read(audio_stream)
    # has multiple channels, do average
    if len(audio_data.shape) == 2:
        audio_data = np.mean(audio_data, axis=1)

    return audio_data


def get_chunk_times(input_fn: str = 'pipe:',
                    input_data: bytes = None,
                    silence_threshold: float = DEFAULT_SILENCE_THRESHOLD,
                    silence_duration: float = DEFAULT_SILENCE_DURATION,
                    start_time: float = None,
                    end_time: float = None):
    _check_input(input_fn, input_data)

    silence_start_re = re.compile(r' silence_start: (?P<start>[0-9]+(\.?[0-9]*))$')
    silence_end_re = re.compile(r' silence_end: (?P<end>[0-9]+(\.?[0-9]*)) ')
    total_duration_re = re.compile(
        r'size=[^ ]+ time=(?P<hours>[0-9]{2}):(?P<minutes>[0-9]{2}):(?P<seconds>[0-9\.]{5}) bitrate=')

    input_kwargs = {}
    if start_time is not None:
        input_kwargs['ss'] = start_time
    else:
        start_time = 0.
    if end_time is not None:
        input_kwargs['t'] = end_time - start_time

    au_filters = [
        'silencedetect=noise={}dB:d={}'.format(silence_threshold,
                                               silence_duration)
    ]

    output_kwargs = {'format': 'null'}
    cmd_args = compile_args(
        input_fn=input_fn,
        input_options=input_kwargs,
        audio_filters=au_filters,
        output_options=output_kwargs)

    stdout, stderr = run_command(
        cmd_args, input=input_data, pipe_stdout=True, pipe_stderr=True)

    lines = stderr.decode().splitlines()

    # Chunks start when silence ends, and chunks end when silence starts.
    chunk_starts = []
    chunk_ends = []
    for line in lines:
        silence_start_match = silence_start_re.search(line)
        silence_end_match = silence_end_re.search(line)
        total_duration_match = total_duration_re.search(line)
        if silence_start_match:
            chunk_ends.append(float(silence_start_match.group('start')))
            if len(chunk_starts) == 0:
                # Started with non-silence.
                chunk_starts.append(start_time)
        elif silence_end_match:
            chunk_starts.append(float(silence_end_match.group('end')))
        elif total_duration_match:
            hours = int(total_duration_match.group('hours'))
            minutes = int(total_duration_match.group('minutes'))
            seconds = float(total_duration_match.group('seconds'))
            end_time = hours * 3600 + minutes * 60 + seconds

    if len(chunk_starts) == 0:
        # No silence found.
        chunk_starts.append(start_time)

    if len(chunk_starts) > len(chunk_ends):
        # Finished with non-silence.
        chunk_ends.append(end_time or 10000000.)

    return list(zip(chunk_starts, chunk_ends))


def split_audio(input_fn: str = 'pipe:',
                input_data: bytes = None,
                silence_threshold=DEFAULT_SILENCE_THRESHOLD,
                silence_duration=DEFAULT_SILENCE_DURATION,
                start_time: float = None,
                end_time: float = None,
                sample_rate: int = 16000,
                verbose=False):
    _check_input(input_fn, input_data)
    chunk_times = get_chunk_times(
        input_fn,
        input_data=input_data,
        silence_threshold=silence_threshold,
        silence_duration=silence_duration,
        start_time=start_time,
        end_time=end_time)
    audio_chunks = list()
    # print("chunk_times", chunk_times)
    for i, (start_time, end_time) in enumerate(chunk_times):
        time = end_time - start_time
        if time < 0:
            continue
        input_kwargs = {
            'ss': start_time,
            't': time
        }

        output_kwargs = {
            'format': 'wav',
            'ar': sample_rate
        }

        cmd_args = compile_args(
            input_fn=input_fn,
            input_options=input_kwargs,
            output_options=output_kwargs)

        stdout, stderr = run_command(
            cmd_args, input=input_data, pipe_stdout=True, pipe_stderr=True)

        audio_stream = io.BytesIO(stdout)
        audio_data, sample_rate = sf.read(audio_stream)
        # has multiple channels, do average
        if len(audio_data.shape) == 2:
            audio_data = np.mean(audio_data, axis=1)
        audio_chunks.append(audio_data)
    return audio_chunks



import subprocess as sp
import tempfile

import numpy as np

from .ffmpeg import compile_args, extract_frame_size
from .helper import _check_input, run_command


def capture_frames(input_fn: str = 'pipe:',
                   input_data: bytes = None,
                   fps: int = None,
                   pix_fmt: str = 'rgb24',
                   vframes: int = -1) -> 'np.ndarray':
    _check_input(input_fn, input_data)

    with tempfile.NamedTemporaryFile(suffix=".gif") as f:
        if input_data:
            f.write(input_data)
            f.flush()
            input_fn = f.name

        video_filters = []
        if fps:
            video_filters += ['fps=%d' % fps]

        output_kwargs = {'format': 'rawvideo', 'pix_fmt': pix_fmt}
        if vframes > 0:
            output_kwargs['vframes'] = vframes

        cmd_args = compile_args(
            input_fn=input_fn,
            video_filters=video_filters,
            output_options=output_kwargs)

        out, err = run_command(cmd_args, pipe_stdout=True, pipe_stderr=True)

        width, height = extract_frame_size(err.decode())

        depth = 3
        if pix_fmt == 'rgba':
            depth = 4

        frames = np.frombuffer(out, np.uint8).copy()
        frames = frames.reshape([-1, height, width, depth])
        return frames


def encode_video(images: 'np.ndarray', frame_rate: int, pix_fmt: str = 'rgb24'):
    cmd = [
        'ffmpeg', '-y', '-f', 'rawvideo', '-vcodec', 'rawvideo', '-r',
        '%.02f' % frame_rate, '-s',
        '%dx%d' % (images[0].shape[1], images[0].shape[0]), '-pix_fmt',
        'rgb24', '-i', '-', '-filter_complex',
        '[0:v]split[x][z];[z]palettegen[y];[x]fifo[x];[x][y]paletteuse', '-r',
        '%.02f' % frame_rate, '-f', 'gif', '-'
    ]
    proc = sp.Popen(cmd, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE)
    for image in images:
        proc.stdin.write(image.tostring())
    out, err = proc.communicate()
    if proc.returncode:
        err = '\n'.join([' '.join(cmd), err.decode('utf8')])
        raise IOError(err)
    del proc
    return out



from typing import List

import numpy as np
from PIL import Image

from ..base import BaseVideoEncoder
from ...helper import batching, get_first_available_gpu


class YouTube8MFeatureExtractor(BaseVideoEncoder):
    """Extracts YouTube8M features for RGB frames.

    First time constructing this class will create directory `yt8m` inside your
    home directory, and will download inception model (85 MB) and YouTube8M PCA
    matrix (15 MB). If you want to use another directory, then pass it to argument
    `model_dir` of constructor.

    If the model_dir exist and contains the necessary files, then files will be
    re-used without download.

    Usage Example:

        from PIL import Image
        import numpy

        # Instantiate extractor. Slow if called first time on your machine, as it
        # needs to download 100 MB.
        extractor = YouTube8MFeatureExtractor()

        image_file = os.path.join(extractor._model_dir, 'cropped_panda.jpg')

        im = numpy.array(Image.open(image_file))
        features = extractor.extract_rgb_frame_features(im)

    ** Note: OpenCV reverses the order of channels (i.e. orders channels as BGR
    instead of RGB). If you are using OpenCV, then you must do:

        im = im[:, :, ::-1]  # Reverses order on last (i.e. channel) dimension.

    then call `extractor.extract_rgb_frame_features(im)`
    """
    batch_size = 64

    def __init__(self, model_dir: str,
                 pca_dir: str,
                 select_layer: str = 'PreLogits',
                 ignore_audio_feature: bool = True,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.model_dir = model_dir
        self.pca_dir = pca_dir
        self.select_layer = select_layer
        self.ignore_audio_feature = ignore_audio_feature
        self.audio_dim = 128
        self.incep_dim = 2048
        self.pca_dim = 1024
        self.inception_size_x = 299
        self.inception_size_y = 299

    def post_init(self):
        import tensorflow as tf
        from .yt8m_feature_extractor_cores.inception_v3 import inception_v3
        from .yt8m_feature_extractor_cores.inception_utils import inception_arg_scope
        import os
        os.environ['CUDA_VISIBLE_DEVICES'] = str(get_first_available_gpu())

        self.pca_mean = np.load(os.path.join(self.pca_dir, 'mean.npy'))[:, 0]
        self.pca_eigenvals = np.load(os.path.join(self.pca_dir, 'eigenvals.npy'))[:self.pca_dim, 0]
        self.pca_eigenvecs = np.load(os.path.join(self.pca_dir, 'eigenvecs.npy')).T[:, :self.pca_dim]

        g = tf.Graph()
        with g.as_default():
            arg_scope = inception_arg_scope()
            inception_v3.default_image_size = self.inception_size_x
            self.inputs = tf.placeholder(tf.float32, (None,
                                                      self.inception_size_x,
                                                      self.inception_size_y, 3))

            with tf.contrib.slim.arg_scope(arg_scope):
                self.logits, self.end_points = inception_v3(self.inputs,
                                                            num_classes=1001,
                                                            is_training=False,
                                                            dropout_keep_prob=1.0)

            config = tf.ConfigProto(log_device_placement=False)
            if self.on_gpu:
                config.gpu_options.allow_growth = True
            self.sess = tf.Session(config=config)
            self.saver = tf.train.Saver()
            self.saver.restore(self.sess, self.model_dir)

    def encode(self, data: List['np.ndarray'], *args, **kwargs) -> List['np.ndarray']:
        video_length = [len(d) for d in data]
        for i in range(1, len(video_length)):
            video_length[i] = video_length[i] + video_length[i - 1]

        data = [(np.array(Image.fromarray(im).resize((self.inception_size_x,
                                                      self.inception_size_y)), dtype=np.float32) * 2 / 255. - 1.)
                for video in data for im in video]

        data = np.stack((list(data[i] for i in range(len(data)))), axis=0)

        @batching
        def _encode(_, data):
            def _pca(data):
                data = np.squeeze(data, axis=(1, 2))
                data = (data - self.pca_mean).reshape((len(data), self.incep_dim))
                data = np.matmul(data, self.pca_eigenvecs)
                data = data / np.sqrt(self.pca_eigenvals + 1e-4)
                return data

            _, end_points_ = self.sess.run((self.logits, self.end_points),
                                           feed_dict={self.inputs: data})
            data = _pca(end_points_[self.select_layer])
            return data

        def _fill_audio_feature(data):
            return list(map(lambda x: np.concatenate((x, np.zeros(shape=(x.shape[0], self.audio_dim))), axis=1), data))

        data = _encode(self, data)

        data = np.split(data, video_length[:-1])

        if self.ignore_audio_feature:
            return _fill_audio_feature(data)
        else:
            return data.astype(np.float32)



import numpy as np

from ..base import BaseNumericEncoder
from ...helper import batching, train_required


class VladEncoder(BaseNumericEncoder):
    batch_size = 2048

    def __init__(self, num_clusters: int,
                 using_faiss_pred: bool = False,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_clusters = num_clusters
        self.using_faiss_pred = using_faiss_pred
        self.centroids = None
        self.index_flat = None

    def kmeans_train(self, vecs):
        import faiss
        kmeans = faiss.Kmeans(vecs.shape[1], self.num_clusters, niter=5, verbose=False)
        kmeans.train(vecs)
        self.centroids = kmeans.centroids
        self.centroids_l2 = np.sum(self.centroids ** 2, axis=1).reshape([1, -1])
        self.centroids_trans = np.transpose(self.centroids)
        if self.using_faiss_pred:
            self.faiss_index()

    def faiss_index(self):
        import faiss
        self.index_flat = faiss.IndexFlatL2(self.centroids.shape[1])
        self.index_flat.add(self.centroids)

    def kmeans_pred(self, vecs):
        if self.using_faiss_pred:
            _, pred = self.index_flat.search(vecs.astype(np.float32), 1)
            return np.reshape(pred, [-1])
        else:
            vecs_l2 = np.sum(vecs ** 2, axis=1).reshape([-1, 1])
            dist = vecs_l2 + self.centroids_l2 - 2 * np.matmul(vecs, self.centroids_trans)
            return np.argmax(dist, axis=-1).reshape([-1]).astype(np.int32)

    @batching
    def train(self, vecs: np.ndarray, *args, **kwargs):
        vecs = vecs.reshape([-1, vecs.shape[-1]])
        assert len(vecs) > self.num_clusters, 'number of data should be larger than number of clusters'
        self.kmeans_train(vecs)

    @train_required
    @batching
    def encode(self, vecs: np.ndarray, *args, **kwargs) -> np.ndarray:
        knn_output = [self.kmeans_pred(vecs_) for vecs_ in vecs]

        output = []
        for chunk_count, chunk in enumerate(vecs):
            res = np.zeros((self.centroids.shape[0], self.centroids.shape[1]))
            for frame_count, frame in enumerate(chunk):
                center_index = knn_output[chunk_count][frame_count]
                res[center_index] += (frame - self.centroids[center_index])
            res = res.reshape([-1])
            output.append(res / np.sum(res ** 2) ** 0.5)

        return np.array(output, dtype=np.float32)

    def _copy_from(self, x: 'VladEncoder') -> None:
        self.num_clusters = x.num_clusters
        self.centroids = x.centroids
        self.centroids_l2 = x.centroids_l2
        self.centroids_trans = np.transpose(self.centroids)
        self.using_faiss_pred = x.using_faiss_pred
        if self.using_faiss_pred:
            self.faiss_index()

    def __setstate__(self, state):
        super().__setstate__(state)
        if self.using_faiss_pred:
            self.faiss_index()

    def __getstate__(self):
        state = super().__getstate__()
        del state['index_flat']
        return state



from typing import Generator

from .base import BaseMapRouter
from ..helper import batch_iterator
from ..proto import gnes_pb2


class BlockRouter(BaseMapRouter):
    """Wait for 'sleep_sec' seconds and forward messages, useful for benchmark"""

    def __init__(self, sleep_sec: int = 5, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sleep_sec = sleep_sec

    def apply(self, msg: 'gnes_pb2.Message', *args, **kwargs):
        import time
        time.sleep(self.sleep_sec)


class PublishRouter(BaseMapRouter):
    """Copy a message 'num_part' time and forward it, useful for PUB-SUB sockets.
    'num_part' is an indicator for downstream sync-barrier, e.g. a ReduceRouter
    """

    def __init__(self, num_part: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_part = num_part

    def apply(self, msg: 'gnes_pb2.Message', *args, **kwargs) -> Generator:
        msg.envelope.num_part.append(self.num_part)


class DocBatchRouter(BaseMapRouter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def apply(self, msg: 'gnes_pb2.Message', *args, **kwargs) -> Generator:
        if self.batch_size and self.batch_size > 0:
            batches = [b for b in batch_iterator(msg.request.index.docs, self.batch_size)]
            num_part = len(batches)
            for p_idx, b in enumerate(batches, start=1):
                _msg = gnes_pb2.Message()
                _msg.CopyFrom(msg)
                _msg.request.index.ClearField('docs')
                _msg.request.index.docs.extend(b)
                _msg.envelope.part_id = p_idx
                _msg.envelope.num_part.append(num_part)
                yield _msg



from . import Flow


class BaseIndexFlow(Flow):
    """
    BaseIndexFlow defines a common service pipeline when indexing.

    It can not be directly used as all services are using the base module by default.
    You have to use :py:meth:`set` to change the `yaml_path` of each service.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        (self.add_preprocessor(name='prep', yaml_path='BasePreprocessor', copy_flow=False)
         .add_encoder(name='enc', yaml_path='BaseEncoder', copy_flow=False)
         .add_indexer(name='vec_idx', yaml_path='BaseIndexer', copy_flow=False)
         .add_indexer(name='doc_idx', yaml_path='BaseIndexer', recv_from='prep', copy_flow=False)
         .add_router(name='sync', yaml_path='BaseReduceRouter',
                     num_part=2, recv_from=['vec_idx', 'doc_idx'], copy_flow=False))


class BaseQueryFlow(Flow):
    """
    BaseIndexFlow defines a common service pipeline when indexing.

    It can not be directly used as all services are using the base module by default.
    You have to use :py:meth:`set` to change the `yaml_path` of each service.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        (self.add_preprocessor(name='prep', yaml_path='BasePreprocessor', copy_flow=False)
         .add_encoder(name='enc', yaml_path='BaseEncoder', copy_flow=False)
         .add_indexer(name='vec_idx', yaml_path='BaseIndexer', copy_flow=False)
         .add_router(name='scorer', yaml_path='Chunk2DocTopkReducer', copy_flow=False)
         .add_indexer(name='doc_idx', yaml_path='BaseIndexer', copy_flow=False))



# A key-value map for Class to the (module)file it located in
from ..base import register_all_class

_cls2file_map = {
    'BaseRouter': 'base',
    'BaseMapRouter': 'base',
    'BaseReduceRouter': 'base',
    'BaseTopkReduceRouter': 'base',
    'BaseEmbedReduceRouter': 'base',
    'DocTopkReducer': 'reduce',
    'ChunkTopkReducer': 'reduce',
    'DocFillReducer': 'reduce',
    'PublishRouter': 'map',
    'DocBatchRouter': 'map',
    'ConcatEmbedRouter': 'reduce',
    'AvgEmbedRouter': 'reduce'
}

register_all_class(_cls2file_map, 'router')



from typing import List, Tuple

import numpy as np

from .base import get_unary_score, CombinedScoreFn


class WeightedChunkScoreFn(CombinedScoreFn):
    """score = d_chunk.weight * relevance * q_chunk.weight"""

    def __call__(self, last_score: 'gnes_pb2.Response.QueryResponse.ScoredResult.Score',
                 q_chunk: 'gnes_pb2.Chunk',
                 d_chunk: 'gnes_pb2.Chunk', *args, **kwargs):
        q_chunk_weight = get_unary_score(value=q_chunk.weight,
                                         name='query chunk weight',
                                         offset=q_chunk.offset)
        d_chunk_weight = get_unary_score(value=d_chunk.weight,
                                         name='document chunk weight',
                                         doc_id=d_chunk.doc_id,
                                         offset=d_chunk.offset)

        return super().__call__(last_score, q_chunk_weight, d_chunk_weight)


class WeightedChunkOffsetScoreFn(CombinedScoreFn):
    """
    score = d_chunk.weight * relevance * offset_divergence * q_chunk.weight
    offset_divergence is calculated based on doc_type:
        TEXT && VIDEO && AUDIO: offset is 1-D
        IMAGE: offset is 2-D
    """

    def __call__(self, last_score: 'gnes_pb2.Response.QueryResponse.ScoredResult.Score',
                 q_chunk: 'gnes_pb2.Chunk',
                 d_chunk: 'gnes_pb2.Chunk', *args, **kwargs):
        q_chunk_weight = get_unary_score(value=q_chunk.weight,
                                         name='query chunk weight',
                                         offset=str(q_chunk.offset))
        d_chunk_weight = get_unary_score(value=d_chunk.weight,
                                         name='document chunk weight',
                                         doc_id=d_chunk.doc_id,
                                         offset=str(d_chunk.offset))
        offset_divergence = get_unary_score(value=self._cal_divergence(q_chunk, d_chunk),
                                            name='offset divergence')
        return super().__call__(last_score, q_chunk_weight, d_chunk_weight, offset_divergence)

    @staticmethod
    def _cal_divergence(q_chunk: 'gnes_pb2.Chunk', d_chunk: 'gnes_pb2.Chunk'):
        if q_chunk.offset_nd and d_chunk.offset_nd:
            return 1 / (1 + np.sqrt((q_chunk.offset_nd[0] - d_chunk.offset_nd[0]) ** 2 +
                                    (q_chunk.offset_nd[1] - d_chunk.offset_nd[1]) ** 2))
        else:
            return np.abs(q_chunk.offset - d_chunk.offset)


class CoordChunkScoreFn(CombinedScoreFn):
    """
    score = relevance * query_coordination
    query_coordination: #chunks return / #chunks in this doc(query doc)
    """

    def __call__(self, last_score: 'gnes_pb2.Response.QueryResponse.ScoredResult.Score',
                 q_chunk: 'gnes_pb2.Chunk',
                 d_chunk: 'gnes_pb2.Chunk',
                 queried_results: List[List[Tuple]],
                 *args, **kwargs):
        query_coordination = get_unary_score(value=self._cal_query_coord(d_chunk, queried_results),
                                             name='query coordination')
        return super().__call__(last_score, query_coordination)

    def _cal_query_coord(self, d_chunk: 'gnes_pb2.Chunk', queried_results: List[List[Tuple]]):
        doc_id = d_chunk.doc_id
        total_chunks = self._context.num_chunks_in_doc(doc_id)
        queried_doc_id, _, _, _ = zip(*(queried_results[0]))
        recall_chunks = queried_doc_id.count(doc_id)
        return recall_chunks / total_chunks


class TFIDFChunkScoreFn(CombinedScoreFn):
    """
    score = relevance * tf(q_chunk) * (idf(q_chunk)**2)
    tf(q_chunk) is calculated based on the relevance of query result.
    tf(q_chunk) = number of queried chunks where relevance >= threshold
    idf(q_chunk) = log(total_chunks / tf(q_chunk) + 1)
    """

    def __init__(self, threshold: float = 0.8, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.threshold = threshold

    def __call__(self, last_score: 'gnes_pb2.Response.QueryResponse.ScoredResult.Score',
                 q_chunk: 'gnes_pb2.Chunk',
                 d_chunk: 'gnes_pb2.Chunk',
                 queried_results: List[List[Tuple]],
                 *args, **kwargs):
        tf_idf = get_unary_score(value=self._cal_tf_idf(queried_results),
                                 name='query tf-idf')
        return super().__call__(last_score, tf_idf)

    def _cal_tf_idf(self, queried_results: List[List[Tuple]]):
        _, _, _, queried_relevance = zip(*(queried_results[0]))
        tf = len(list(filter(lambda x: x >= self.threshold, queried_relevance)))

        total_chunks = self._context.num_chunks
        idf = np.log10(total_chunks / (tf + 1))
        return tf * (idf ** 2)


class BM25ChunkScoreFn(CombinedScoreFn):
    """
    score = relevance * idf(q_chunk) * tf(q_chunk) * (k1 + 1) / (tf(q_chunk) +
                            k1 * (1 - b + b * (chunk_in_doc / avg_chunk_in_doc)))

    in bm25 algorithm:
             idf(q_chunk) = log(1 + (doc_count - f(q_chunk) +0.5) / (f(q_chunk) + 0.5)),
    where f(q_chunk) is number of docs that contains q_chunk. In our system, this denotes number of docs
    appearing in query results.

    In elastic search, b = 0.75, k1 = 1.2
    """

    def __init__(self, threshold: float = 0.8, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.threshold = threshold
        self.k1 = 1.2
        self.b = 0.75

    def __call__(self, last_score: 'gnes_pb2.Response.QueryResponse.ScoredResult.Score',
                 q_chunk: 'gnes_pb2.Chunk',
                 d_chunk: 'gnes_pb2.Chunk',
                 queried_results: List[List[Tuple]],
                 *args, **kwargs):
        bm25 = get_unary_score(value=self._cal_bm25(d_chunk, queried_results),
                               name='query bm25')
        return super().__call__(last_score, bm25)

    def _cal_bm25(self, d_chunk: 'gnes_pb2.Chunk', queried_results: List[List[Tuple]]):
        doc_id = d_chunk.doc_id
        _, _, _, queried_relevance = zip(*(queried_results[0]))
        tf = len(list(filter(lambda x: x >= self.threshold, queried_relevance)))

        total_chunks = self._context.num_chunks
        idf = np.log10(1 + (total_chunks - tf + 0.5) / (tf + 0.5))
        return idf * tf * (self.k1 + 1) / (tf + self.k1 * (1 - self.b + self.b *
                                                           (self._context.num_chunks_in_doc(
                                                               doc_id) * self._context.num_docs / self._context.num_chunks)))





import os
import unittest

from gnes.cli.parser import set_preprocessor_parser, _set_client_parser
from gnes.client.base import ZmqClient
from gnes.proto import gnes_pb2, RequestGenerator, blob2array
from gnes.service.base import ServiceManager
from gnes.service.preprocessor import PreprocessorService


class TestVideoDecode(unittest.TestCase):

    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.yml_path = os.path.join(self.dirname, 'yaml', 'preprocessor-video_decode.yml')
        self.video_path = os.path.join(self.dirname, 'videos')

    def test_video_decode_preprocessor(self):
        args = set_preprocessor_parser().parse_args(['--yaml_path', self.yml_path])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)])
        video_bytes = [
            open(os.path.join(self.video_path, _), 'rb').read()
            for _ in os.listdir(self.video_path)
        ]

        with ServiceManager(PreprocessorService, args), ZmqClient(c_args) as client:
            for req in RequestGenerator.index(video_bytes):
                msg = gnes_pb2.Message()
                msg.request.index.CopyFrom(req.index)
                client.send_message(msg)
                r = client.recv_message()
                for d in r.request.index.docs:
                    self.assertGreater(len(d.chunks), 0)
                    for _ in range(len(d.chunks)):
                        shape = blob2array(d.chunks[_].blob).shape
                        self.assertEqual(shape, (299, 299, 3))



from typing import List


def encode_video(images: List['np.ndarray'], frame_rate: int, pix_fmt: str = 'rgb24'):
    import webp
    from PIL import Image

    height, width, _ = images[0].shape
    if pix_fmt == 'rgb24':
        pix_fmt = 'RGB'
    # Save an animation
    enc = webp.WebPAnimEncoder.new(width, height)
    timestamp_ms = 0
    duration = 1000 // frame_rate
    for x in images:
        img = Image.fromarray(x.copy(), pix_fmt)
        pic = webp.WebPPicture.from_pil(img)
        enc.encode_frame(pic, timestamp_ms)
        timestamp_ms += duration
    anim_data = enc.assemble(timestamp_ms)
    return bytes(anim_data.buffer())



# A key-value map for Class to the (module)file it located in
from ..base import register_all_class

_cls2file_map = {
    'BaseScoreFn': 'base',
    'CombinedScoreFn': 'base',
    'ModifierScoreFn': 'base',
    'WeightedChunkScoreFn': 'chunk',
    'WeightedDocScoreFn': 'doc',
    'Normalizer1': 'normalize',
    'Normalizer2': 'normalize',
    'Normalizer3': 'normalize',
    'Normalizer4': 'normalize',
    'Normalizer5': 'normalize',
}

register_all_class(_cls2file_map, 'score_fn')



import math
import numpy as np

from gnes.preprocessor.base import BaseVideoPreprocessor
from gnes.proto import gnes_pb2, array2blob, blob2array


class FrameSelectPreprocessor(BaseVideoPreprocessor):

    def __init__(self,
                 sframes: int = 1,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.sframes = sframes

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)
        if len(doc.chunks) > 0:
            for chunk in doc.chunks:
                images = blob2array(chunk.blob)
                if len(images) == 0:
                    self.logger.warning("this chunk has no frame!")
                elif self.sframes == 1:
                    idx = [int(len(images) / 2)]
                    chunk.blob.CopyFrom(array2blob(images[idx]))
                elif self.sframes > 0 and len(images) > self.sframes:
                    if len(images) >= 2 * self.sframes:
                        step = math.ceil(len(images) / self.sframes)
                        chunk.blob.CopyFrom(array2blob(images[::step]))
                    else:
                        idx = np.sort(np.random.choice(len(images), self.sframes, replace=False))
                        chunk.blob.CopyFrom(array2blob(images[idx]))
                del images
        else:
            self.logger.error(
                'bad document: "doc.chunks" is empty!')



# A key-value map for Class to the (module)file it located in
from ..base import register_all_class

_cls2file_map = {
    'BasePreprocessor': 'base',
    'PipelinePreprocessor': 'base',
    'SentSplitPreprocessor': 'text.split',
    'BaseImagePreprocessor': 'base',
    'BaseTextPreprocessor': 'base',
    'VanillaSlidingPreprocessor': 'image.sliding_window',
    'WeightedSlidingPreprocessor': 'image.sliding_window',
    'SegmentPreprocessor': 'image.segmentation',
    'UnaryPreprocessor': 'base',
    'ResizeChunkPreprocessor': 'image.resize',
    'BaseVideoPreprocessor': 'base',
    'FFmpegPreprocessor': 'video.ffmpeg',
    'FFmpegVideoSegmentor': 'video.ffmpeg',
    'ShotDetectorPreprocessor': 'video.shot_detector',
    'VideoEncoderPreprocessor': 'video.video_encoder',
    'VideoDecoderPreprocessor': 'video.video_decoder',
    'AudioVanilla': 'audio.audio_vanilla',
    'BaseAudioPreprocessor': 'base',
    'RawChunkPreprocessor': 'base',
    'GifChunkPreprocessor': 'video.ffmpeg',
    'VggishPreprocessor': 'audio.vggish_example',
    'FrameSelectPreprocessor': 'video.frame_select'
}

register_all_class(_cls2file_map, 'preprocessor')



# do not change this line manually
# this is managed by git tag and updated on every release
__version__ = '0.0.47'

# do not change this line manually
# this is managed by shell/make-proto.sh and updated on every execution
__proto_version__ = '0.0.10'



import io

import numpy as np

from ..base import TrainableBase, CompositionalTrainableBase
from ..proto import gnes_pb2, array2blob


class BasePreprocessor(TrainableBase):
    doc_type = gnes_pb2.Document.UNKNOWN

    def __init__(self,
                 uniform_doc_weight: bool = True,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.uniform_doc_weight = uniform_doc_weight

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        doc.doc_type = self.doc_type
        if not doc.weight and self.uniform_doc_weight:
            doc.weight = 1.0


class BaseTextPreprocessor(BasePreprocessor):
    doc_type = gnes_pb2.Document.TEXT


class BaseAudioPreprocessor(BasePreprocessor):
    doc_type = gnes_pb2.Document.AUDIO


class BaseImagePreprocessor(BasePreprocessor):
    doc_type = gnes_pb2.Document.IMAGE


class BaseVideoPreprocessor(BasePreprocessor):
    doc_type = gnes_pb2.Document.VIDEO


class PipelinePreprocessor(CompositionalTrainableBase):
    def apply(self, doc: 'gnes_pb2.Document') -> None:
        if not self.components:
            raise NotImplementedError
        for be in self.components:
            be.apply(doc)

    def train(self, data, *args, **kwargs):
        if not self.components:
            raise NotImplementedError
        for idx, be in enumerate(self.components):
            be.train(data, *args, **kwargs)
            if idx + 1 < len(self.components):
                data = be.apply(data, *args, **kwargs)


class UnaryPreprocessor(BasePreprocessor):
    is_trained = True

    def __init__(self, doc_type: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.doc_type = doc_type

    def apply(self, doc: 'gnes_pb2.Document'):
        super().apply(doc)
        c = doc.chunks.add()
        c.doc_id = doc.doc_id
        c.offset = 0
        c.weight = 1.
        if doc.raw_bytes:
            self.raw_to_chunk(c, doc.raw_bytes)
        else:
            self.logger.error('bad document: "raw_bytes" is empty!')

    def raw_to_chunk(self, chunk: 'gnes_pb2.Chunk', raw_bytes: bytes):
        if self.doc_type == gnes_pb2.Document.TEXT:
            chunk.text = raw_bytes.decode()
        elif self.doc_type == gnes_pb2.Document.IMAGE:
            from PIL import Image
            img = np.array(Image.open(io.BytesIO(raw_bytes)))
            chunk.blob.CopyFrom(array2blob(img))
        elif self.doc_type == gnes_pb2.Document.VIDEO:
            raise NotImplementedError
        else:
            raise NotImplementedError


class RawChunkPreprocessor(BasePreprocessor):

    @staticmethod
    def _parse_chunk(chunk: 'gnes_pb2.Chunk', *args, **kwargs):
        raise NotImplementedError

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        if doc.raw_bytes:
            for chunk in doc.chunks:
                chunk.raw = self._parse_chunk(chunk)
        else:
            self.logger.error('bad document: "raw_bytes" is empty!')



import grpc

from .base import BaseService as BS, MessageHandler
from ..helper import PathImporter
from ..proto import gnes_pb2


class GRPCService(BS):
    handler = MessageHandler(BS.handler)

    def post_init(self):
        self.channel = grpc.insecure_channel(
            '%s:%s' % (self.args.grpc_host, self.args.grpc_port),
            options=[('grpc.max_send_message_length', self.args.max_message_size),
                     ('grpc.max_receive_message_length', self.args.max_message_size)])

        m = PathImporter.add_modules(self.args.pb2_path, self.args.pb2_grpc_path)

        # build stub
        self.stub = getattr(m, self.args.stub_name)(self.channel)

    def close(self):
        self.channel.close()
        super().close()

    @handler.register(NotImplementedError)
    def _handler_default(self, msg: 'gnes_pb2.Message'):
        yield getattr(self.stub, self.args.api_name)(msg)



from typing import List

import numpy as np

from ..base import BaseAudioEncoder
from ...helper import batching, get_first_available_gpu


class VggishEncoder(BaseAudioEncoder):
    def __init__(self, model_dir: str,
                 max_length: int = 10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_dir = model_dir
        self.max_length = max_length

    def post_init(self):
        import os
        import tensorflow as tf
        from .vggish_cores import vggish_slim
        from .vggish_cores import vggish_params
        from .vggish_cores import vggish_postprocess

        os.environ['CUDA_VISIBLE_DEVICES'] = str(get_first_available_gpu())

        self.graph = tf.Graph()

        with tf.Graph().as_default():
            self._sess = tf.Session()
            vggish_slim.define_vggish_slim(training=False)
            vggish_slim.load_vggish_slim_checkpoint(self._sess, self.model_dir + "/vggish_model.ckpt")
            self._audio_inputs = self._sess.graph.get_tensor_by_name(vggish_params.INPUT_TENSOR_NAME)
            self._embedding = self._sess.graph.get_tensor_by_name(vggish_params.OUTPUT_TENSOR_NAME)
        self._pproc = vggish_postprocess.Postprocessor(self.model_dir + "/vggish_pca_params.npz")

    @batching
    def encode(self, audio: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        # used for np.split()
        audio_length = [len(au) for au in audio]
        for i in range(1, len(audio_length)):
            audio_length[i] = audio_length[i] + audio_length[i - 1]

        # concat at axis = 0 to make sure the dimension is: (#vggish_example, D)
        audio_ = np.concatenate((list(audio[i] for i in range(len(audio)))), axis=0)

        [embedding_batch] = self._sess.run([self._embedding], feed_dict={self._audio_inputs: audio_})
        postprocessed_batch = self._pproc.postprocess(embedding_batch)

        audio_features = np.split(postprocessed_batch, audio_length[:-1])
        return np.array([features.mean(axis=0) for features in audio_features])



# ==============================================================================

"""Global parameters for the VGGish model.

See vggish_slim.py for more information.
"""

# Architectural constants.
NUM_FRAMES = 96  # Frames in input mel-spectrogram patch.
NUM_BANDS = 64  # Frequency bands in input mel-spectrogram patch.
EMBEDDING_SIZE = 128  # Size of embedding layer.

# Parameters used for embedding postprocessing.
PCA_EIGEN_VECTORS_NAME = 'pca_eigen_vectors'
PCA_MEANS_NAME = 'pca_means'
QUANTIZE_MIN_VAL = -2.0
QUANTIZE_MAX_VAL = +2.0

# Hyperparameters used in training.
INIT_STDDEV = 0.01  # Standard deviation used to initialize weights.
LEARNING_RATE = 1e-4  # Learning rate for the Adam optimizer.
ADAM_EPSILON = 1e-8  # Epsilon for the Adam optimizer.

# Names of ops, tensors, and features.
INPUT_OP_NAME = 'vggish/input_features'
INPUT_TENSOR_NAME = INPUT_OP_NAME + ':0'
OUTPUT_OP_NAME = 'vggish/embedding'
OUTPUT_TENSOR_NAME = OUTPUT_OP_NAME + ':0'
AUDIO_EMBEDDING_FEATURE_NAME = 'audio_embedding'



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the definition for inception v3 classification network."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from . import inception_utils

slim = tf.contrib.slim
trunc_normal = lambda stddev: tf.truncated_normal_initializer(0.0, stddev)


def inception_v3_base(inputs,
                      final_endpoint='Mixed_7c',
                      min_depth=16,
                      depth_multiplier=1.0,
                      scope=None):
    """Inception model from http://arxiv.org/abs/1512.00567.
    Constructs an Inception v3 network from inputs to the given final endpoint.
    This method can construct the network up to the final inception block
    Mixed_7c.
    Note that the names of the layers in the paper do not correspond to the names
    of the endpoints registered by this function although they build the same
    network.
    Here is a mapping from the old_names to the new names:
    Old name          | New name
    =======================================
    conv0             | Conv2d_1a_3x3
    conv1             | Conv2d_2a_3x3
    conv2             | Conv2d_2b_3x3
    pool1             | MaxPool_3a_3x3
    conv3             | Conv2d_3b_1x1
    conv4             | Conv2d_4a_3x3
    pool2             | MaxPool_5a_3x3
    mixed_35x35x256a  | Mixed_5b
    mixed_35x35x288a  | Mixed_5c
    mixed_35x35x288b  | Mixed_5d
    mixed_17x17x768a  | Mixed_6a
    mixed_17x17x768b  | Mixed_6b
    mixed_17x17x768c  | Mixed_6c
    mixed_17x17x768d  | Mixed_6d
    mixed_17x17x768e  | Mixed_6e
    mixed_8x8x1280a   | Mixed_7a
    mixed_8x8x2048a   | Mixed_7b
    mixed_8x8x2048b   | Mixed_7c
    Args:
      inputs: a tensor of size [batch_size, height, width, channels].
      final_endpoint: specifies the endpoint to construct the network up to. It
        can be one of ['Conv2d_1a_3x3', 'Conv2d_2a_3x3', 'Conv2d_2b_3x3',
        'MaxPool_3a_3x3', 'Conv2d_3b_1x1', 'Conv2d_4a_3x3', 'MaxPool_5a_3x3',
        'Mixed_5b', 'Mixed_5c', 'Mixed_5d', 'Mixed_6a', 'Mixed_6b', 'Mixed_6c',
        'Mixed_6d', 'Mixed_6e', 'Mixed_7a', 'Mixed_7b', 'Mixed_7c'].
      min_depth: Minimum depth value (number of channels) for all convolution ops.
        Enforced when depth_multiplier < 1, and not an active constraint when
        depth_multiplier >= 1.
      depth_multiplier: Float multiplier for the depth (number of channels)
        for all convolution ops. The value must be greater than zero. Typical
        usage will be to set this value in (0, 1) to reduce the number of
        parameters or computation cost of the model.
      scope: Optional variable_scope.
    Returns:
      tensor_out: output tensor corresponding to the final_endpoint.
      end_points: a set of activations for external use, for example summaries or
                  losses.
    Raises:
      ValueError: if final_endpoint is not set to one of the predefined values,
                  or depth_multiplier <= 0
    """
    # end_points will collect relevant activations for external use, for example
    # summaries or losses.
    end_points = {}

    if depth_multiplier <= 0:
        raise ValueError('depth_multiplier is not greater than zero.')
    depth = lambda d: max(int(d * depth_multiplier), min_depth)

    with tf.variable_scope(scope, 'InceptionV3', [inputs]):
        with slim.arg_scope([slim.conv2d, slim.max_pool2d, slim.avg_pool2d],
                            stride=1, padding='VALID'):
            # 299 x 299 x 3
            end_point = 'Conv2d_1a_3x3'
            net = slim.conv2d(inputs, depth(32), [3, 3], stride=2, scope=end_point)
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points
            # 149 x 149 x 32
            end_point = 'Conv2d_2a_3x3'
            net = slim.conv2d(net, depth(32), [3, 3], scope=end_point)
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points
            # 147 x 147 x 32
            end_point = 'Conv2d_2b_3x3'
            net = slim.conv2d(net, depth(64), [3, 3], padding='SAME', scope=end_point)
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points
            # 147 x 147 x 64
            end_point = 'MaxPool_3a_3x3'
            net = slim.max_pool2d(net, [3, 3], stride=2, scope=end_point)
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points
            # 73 x 73 x 64
            end_point = 'Conv2d_3b_1x1'
            net = slim.conv2d(net, depth(80), [1, 1], scope=end_point)
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points
            # 73 x 73 x 80.
            end_point = 'Conv2d_4a_3x3'
            net = slim.conv2d(net, depth(192), [3, 3], scope=end_point)
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points
            # 71 x 71 x 192.
            end_point = 'MaxPool_5a_3x3'
            net = slim.max_pool2d(net, [3, 3], stride=2, scope=end_point)
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points
            # 35 x 35 x 192.

        # Inception blocks
        with slim.arg_scope([slim.conv2d, slim.max_pool2d, slim.avg_pool2d],
                            stride=1, padding='SAME'):
            # mixed: 35 x 35 x 256.
            end_point = 'Mixed_5b'
            with tf.variable_scope(end_point):
                with tf.variable_scope('Branch_0'):
                    branch_0 = slim.conv2d(net, depth(64), [1, 1], scope='Conv2d_0a_1x1')
                with tf.variable_scope('Branch_1'):
                    branch_1 = slim.conv2d(net, depth(48), [1, 1], scope='Conv2d_0a_1x1')
                    branch_1 = slim.conv2d(branch_1, depth(64), [5, 5],
                                           scope='Conv2d_0b_5x5')
                with tf.variable_scope('Branch_2'):
                    branch_2 = slim.conv2d(net, depth(64), [1, 1], scope='Conv2d_0a_1x1')
                    branch_2 = slim.conv2d(branch_2, depth(96), [3, 3],
                                           scope='Conv2d_0b_3x3')
                    branch_2 = slim.conv2d(branch_2, depth(96), [3, 3],
                                           scope='Conv2d_0c_3x3')
                with tf.variable_scope('Branch_3'):
                    branch_3 = slim.avg_pool2d(net, [3, 3], scope='AvgPool_0a_3x3')
                    branch_3 = slim.conv2d(branch_3, depth(32), [1, 1],
                                           scope='Conv2d_0b_1x1')
                net = tf.concat(axis=3, values=[branch_0, branch_1, branch_2, branch_3])
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points

            # mixed_1: 35 x 35 x 288.
            end_point = 'Mixed_5c'
            with tf.variable_scope(end_point):
                with tf.variable_scope('Branch_0'):
                    branch_0 = slim.conv2d(net, depth(64), [1, 1], scope='Conv2d_0a_1x1')
                with tf.variable_scope('Branch_1'):
                    branch_1 = slim.conv2d(net, depth(48), [1, 1], scope='Conv2d_0b_1x1')
                    branch_1 = slim.conv2d(branch_1, depth(64), [5, 5],
                                           scope='Conv_1_0c_5x5')
                with tf.variable_scope('Branch_2'):
                    branch_2 = slim.conv2d(net, depth(64), [1, 1],
                                           scope='Conv2d_0a_1x1')
                    branch_2 = slim.conv2d(branch_2, depth(96), [3, 3],
                                           scope='Conv2d_0b_3x3')
                    branch_2 = slim.conv2d(branch_2, depth(96), [3, 3],
                                           scope='Conv2d_0c_3x3')
                with tf.variable_scope('Branch_3'):
                    branch_3 = slim.avg_pool2d(net, [3, 3], scope='AvgPool_0a_3x3')
                    branch_3 = slim.conv2d(branch_3, depth(64), [1, 1],
                                           scope='Conv2d_0b_1x1')
                net = tf.concat(axis=3, values=[branch_0, branch_1, branch_2, branch_3])
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points

            # mixed_2: 35 x 35 x 288.
            end_point = 'Mixed_5d'
            with tf.variable_scope(end_point):
                with tf.variable_scope('Branch_0'):
                    branch_0 = slim.conv2d(net, depth(64), [1, 1], scope='Conv2d_0a_1x1')
                with tf.variable_scope('Branch_1'):
                    branch_1 = slim.conv2d(net, depth(48), [1, 1], scope='Conv2d_0a_1x1')
                    branch_1 = slim.conv2d(branch_1, depth(64), [5, 5],
                                           scope='Conv2d_0b_5x5')
                with tf.variable_scope('Branch_2'):
                    branch_2 = slim.conv2d(net, depth(64), [1, 1], scope='Conv2d_0a_1x1')
                    branch_2 = slim.conv2d(branch_2, depth(96), [3, 3],
                                           scope='Conv2d_0b_3x3')
                    branch_2 = slim.conv2d(branch_2, depth(96), [3, 3],
                                           scope='Conv2d_0c_3x3')
                with tf.variable_scope('Branch_3'):
                    branch_3 = slim.avg_pool2d(net, [3, 3], scope='AvgPool_0a_3x3')
                    branch_3 = slim.conv2d(branch_3, depth(64), [1, 1],
                                           scope='Conv2d_0b_1x1')
                net = tf.concat(axis=3, values=[branch_0, branch_1, branch_2, branch_3])
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points

            # mixed_3: 17 x 17 x 768.
            end_point = 'Mixed_6a'
            with tf.variable_scope(end_point):
                with tf.variable_scope('Branch_0'):
                    branch_0 = slim.conv2d(net, depth(384), [3, 3], stride=2,
                                           padding='VALID', scope='Conv2d_1a_1x1')
                with tf.variable_scope('Branch_1'):
                    branch_1 = slim.conv2d(net, depth(64), [1, 1], scope='Conv2d_0a_1x1')
                    branch_1 = slim.conv2d(branch_1, depth(96), [3, 3],
                                           scope='Conv2d_0b_3x3')
                    branch_1 = slim.conv2d(branch_1, depth(96), [3, 3], stride=2,
                                           padding='VALID', scope='Conv2d_1a_1x1')
                with tf.variable_scope('Branch_2'):
                    branch_2 = slim.max_pool2d(net, [3, 3], stride=2, padding='VALID',
                                               scope='MaxPool_1a_3x3')
                net = tf.concat(axis=3, values=[branch_0, branch_1, branch_2])
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points

            # mixed4: 17 x 17 x 768.
            end_point = 'Mixed_6b'
            with tf.variable_scope(end_point):
                with tf.variable_scope('Branch_0'):
                    branch_0 = slim.conv2d(net, depth(192), [1, 1], scope='Conv2d_0a_1x1')
                with tf.variable_scope('Branch_1'):
                    branch_1 = slim.conv2d(net, depth(128), [1, 1], scope='Conv2d_0a_1x1')
                    branch_1 = slim.conv2d(branch_1, depth(128), [1, 7],
                                           scope='Conv2d_0b_1x7')
                    branch_1 = slim.conv2d(branch_1, depth(192), [7, 1],
                                           scope='Conv2d_0c_7x1')
                with tf.variable_scope('Branch_2'):
                    branch_2 = slim.conv2d(net, depth(128), [1, 1], scope='Conv2d_0a_1x1')
                    branch_2 = slim.conv2d(branch_2, depth(128), [7, 1],
                                           scope='Conv2d_0b_7x1')
                    branch_2 = slim.conv2d(branch_2, depth(128), [1, 7],
                                           scope='Conv2d_0c_1x7')
                    branch_2 = slim.conv2d(branch_2, depth(128), [7, 1],
                                           scope='Conv2d_0d_7x1')
                    branch_2 = slim.conv2d(branch_2, depth(192), [1, 7],
                                           scope='Conv2d_0e_1x7')
                with tf.variable_scope('Branch_3'):
                    branch_3 = slim.avg_pool2d(net, [3, 3], scope='AvgPool_0a_3x3')
                    branch_3 = slim.conv2d(branch_3, depth(192), [1, 1],
                                           scope='Conv2d_0b_1x1')
                net = tf.concat(axis=3, values=[branch_0, branch_1, branch_2, branch_3])
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points

            # mixed_5: 17 x 17 x 768.
            end_point = 'Mixed_6c'
            with tf.variable_scope(end_point):
                with tf.variable_scope('Branch_0'):
                    branch_0 = slim.conv2d(net, depth(192), [1, 1], scope='Conv2d_0a_1x1')
                with tf.variable_scope('Branch_1'):
                    branch_1 = slim.conv2d(net, depth(160), [1, 1], scope='Conv2d_0a_1x1')
                    branch_1 = slim.conv2d(branch_1, depth(160), [1, 7],
                                           scope='Conv2d_0b_1x7')
                    branch_1 = slim.conv2d(branch_1, depth(192), [7, 1],
                                           scope='Conv2d_0c_7x1')
                with tf.variable_scope('Branch_2'):
                    branch_2 = slim.conv2d(net, depth(160), [1, 1], scope='Conv2d_0a_1x1')
                    branch_2 = slim.conv2d(branch_2, depth(160), [7, 1],
                                           scope='Conv2d_0b_7x1')
                    branch_2 = slim.conv2d(branch_2, depth(160), [1, 7],
                                           scope='Conv2d_0c_1x7')
                    branch_2 = slim.conv2d(branch_2, depth(160), [7, 1],
                                           scope='Conv2d_0d_7x1')
                    branch_2 = slim.conv2d(branch_2, depth(192), [1, 7],
                                           scope='Conv2d_0e_1x7')
                with tf.variable_scope('Branch_3'):
                    branch_3 = slim.avg_pool2d(net, [3, 3], scope='AvgPool_0a_3x3')
                    branch_3 = slim.conv2d(branch_3, depth(192), [1, 1],
                                           scope='Conv2d_0b_1x1')
                net = tf.concat(axis=3, values=[branch_0, branch_1, branch_2, branch_3])
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points
            # mixed_6: 17 x 17 x 768.
            end_point = 'Mixed_6d'
            with tf.variable_scope(end_point):
                with tf.variable_scope('Branch_0'):
                    branch_0 = slim.conv2d(net, depth(192), [1, 1], scope='Conv2d_0a_1x1')
                with tf.variable_scope('Branch_1'):
                    branch_1 = slim.conv2d(net, depth(160), [1, 1], scope='Conv2d_0a_1x1')
                    branch_1 = slim.conv2d(branch_1, depth(160), [1, 7],
                                           scope='Conv2d_0b_1x7')
                    branch_1 = slim.conv2d(branch_1, depth(192), [7, 1],
                                           scope='Conv2d_0c_7x1')
                with tf.variable_scope('Branch_2'):
                    branch_2 = slim.conv2d(net, depth(160), [1, 1], scope='Conv2d_0a_1x1')
                    branch_2 = slim.conv2d(branch_2, depth(160), [7, 1],
                                           scope='Conv2d_0b_7x1')
                    branch_2 = slim.conv2d(branch_2, depth(160), [1, 7],
                                           scope='Conv2d_0c_1x7')
                    branch_2 = slim.conv2d(branch_2, depth(160), [7, 1],
                                           scope='Conv2d_0d_7x1')
                    branch_2 = slim.conv2d(branch_2, depth(192), [1, 7],
                                           scope='Conv2d_0e_1x7')
                with tf.variable_scope('Branch_3'):
                    branch_3 = slim.avg_pool2d(net, [3, 3], scope='AvgPool_0a_3x3')
                    branch_3 = slim.conv2d(branch_3, depth(192), [1, 1],
                                           scope='Conv2d_0b_1x1')
                net = tf.concat(axis=3, values=[branch_0, branch_1, branch_2, branch_3])
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points

            # mixed_7: 17 x 17 x 768.
            end_point = 'Mixed_6e'
            with tf.variable_scope(end_point):
                with tf.variable_scope('Branch_0'):
                    branch_0 = slim.conv2d(net, depth(192), [1, 1], scope='Conv2d_0a_1x1')
                with tf.variable_scope('Branch_1'):
                    branch_1 = slim.conv2d(net, depth(192), [1, 1], scope='Conv2d_0a_1x1')
                    branch_1 = slim.conv2d(branch_1, depth(192), [1, 7],
                                           scope='Conv2d_0b_1x7')
                    branch_1 = slim.conv2d(branch_1, depth(192), [7, 1],
                                           scope='Conv2d_0c_7x1')
                with tf.variable_scope('Branch_2'):
                    branch_2 = slim.conv2d(net, depth(192), [1, 1], scope='Conv2d_0a_1x1')
                    branch_2 = slim.conv2d(branch_2, depth(192), [7, 1],
                                           scope='Conv2d_0b_7x1')
                    branch_2 = slim.conv2d(branch_2, depth(192), [1, 7],
                                           scope='Conv2d_0c_1x7')
                    branch_2 = slim.conv2d(branch_2, depth(192), [7, 1],
                                           scope='Conv2d_0d_7x1')
                    branch_2 = slim.conv2d(branch_2, depth(192), [1, 7],
                                           scope='Conv2d_0e_1x7')
                with tf.variable_scope('Branch_3'):
                    branch_3 = slim.avg_pool2d(net, [3, 3], scope='AvgPool_0a_3x3')
                    branch_3 = slim.conv2d(branch_3, depth(192), [1, 1],
                                           scope='Conv2d_0b_1x1')
                net = tf.concat(axis=3, values=[branch_0, branch_1, branch_2, branch_3])
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points

            # mixed_8: 8 x 8 x 1280.
            end_point = 'Mixed_7a'
            with tf.variable_scope(end_point):
                with tf.variable_scope('Branch_0'):
                    branch_0 = slim.conv2d(net, depth(192), [1, 1], scope='Conv2d_0a_1x1')
                    branch_0 = slim.conv2d(branch_0, depth(320), [3, 3], stride=2,
                                           padding='VALID', scope='Conv2d_1a_3x3')
                with tf.variable_scope('Branch_1'):
                    branch_1 = slim.conv2d(net, depth(192), [1, 1], scope='Conv2d_0a_1x1')
                    branch_1 = slim.conv2d(branch_1, depth(192), [1, 7],
                                           scope='Conv2d_0b_1x7')
                    branch_1 = slim.conv2d(branch_1, depth(192), [7, 1],
                                           scope='Conv2d_0c_7x1')
                    branch_1 = slim.conv2d(branch_1, depth(192), [3, 3], stride=2,
                                           padding='VALID', scope='Conv2d_1a_3x3')
                with tf.variable_scope('Branch_2'):
                    branch_2 = slim.max_pool2d(net, [3, 3], stride=2, padding='VALID',
                                               scope='MaxPool_1a_3x3')
                net = tf.concat(axis=3, values=[branch_0, branch_1, branch_2])
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points
            # mixed_9: 8 x 8 x 2048.
            end_point = 'Mixed_7b'
            with tf.variable_scope(end_point):
                with tf.variable_scope('Branch_0'):
                    branch_0 = slim.conv2d(net, depth(320), [1, 1], scope='Conv2d_0a_1x1')
                with tf.variable_scope('Branch_1'):
                    branch_1 = slim.conv2d(net, depth(384), [1, 1], scope='Conv2d_0a_1x1')
                    branch_1 = tf.concat(axis=3, values=[
                        slim.conv2d(branch_1, depth(384), [1, 3], scope='Conv2d_0b_1x3'),
                        slim.conv2d(branch_1, depth(384), [3, 1], scope='Conv2d_0b_3x1')])
                with tf.variable_scope('Branch_2'):
                    branch_2 = slim.conv2d(net, depth(448), [1, 1], scope='Conv2d_0a_1x1')
                    branch_2 = slim.conv2d(
                        branch_2, depth(384), [3, 3], scope='Conv2d_0b_3x3')
                    branch_2 = tf.concat(axis=3, values=[
                        slim.conv2d(branch_2, depth(384), [1, 3], scope='Conv2d_0c_1x3'),
                        slim.conv2d(branch_2, depth(384), [3, 1], scope='Conv2d_0d_3x1')])
                with tf.variable_scope('Branch_3'):
                    branch_3 = slim.avg_pool2d(net, [3, 3], scope='AvgPool_0a_3x3')
                    branch_3 = slim.conv2d(
                        branch_3, depth(192), [1, 1], scope='Conv2d_0b_1x1')
                net = tf.concat(axis=3, values=[branch_0, branch_1, branch_2, branch_3])
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points

            # mixed_10: 8 x 8 x 2048.
            end_point = 'Mixed_7c'
            with tf.variable_scope(end_point):
                with tf.variable_scope('Branch_0'):
                    branch_0 = slim.conv2d(net, depth(320), [1, 1], scope='Conv2d_0a_1x1')
                with tf.variable_scope('Branch_1'):
                    branch_1 = slim.conv2d(net, depth(384), [1, 1], scope='Conv2d_0a_1x1')
                    branch_1 = tf.concat(axis=3, values=[
                        slim.conv2d(branch_1, depth(384), [1, 3], scope='Conv2d_0b_1x3'),
                        slim.conv2d(branch_1, depth(384), [3, 1], scope='Conv2d_0c_3x1')])
                with tf.variable_scope('Branch_2'):
                    branch_2 = slim.conv2d(net, depth(448), [1, 1], scope='Conv2d_0a_1x1')
                    branch_2 = slim.conv2d(
                        branch_2, depth(384), [3, 3], scope='Conv2d_0b_3x3')
                    branch_2 = tf.concat(axis=3, values=[
                        slim.conv2d(branch_2, depth(384), [1, 3], scope='Conv2d_0c_1x3'),
                        slim.conv2d(branch_2, depth(384), [3, 1], scope='Conv2d_0d_3x1')])
                with tf.variable_scope('Branch_3'):
                    branch_3 = slim.avg_pool2d(net, [3, 3], scope='AvgPool_0a_3x3')
                    branch_3 = slim.conv2d(
                        branch_3, depth(192), [1, 1], scope='Conv2d_0b_1x1')
                net = tf.concat(axis=3, values=[branch_0, branch_1, branch_2, branch_3])
            end_points[end_point] = net
            if end_point == final_endpoint: return net, end_points
        raise ValueError('Unknown final endpoint %s' % final_endpoint)


def inception_v3(inputs,
                 num_classes=1000,
                 is_training=True,
                 dropout_keep_prob=0.8,
                 min_depth=16,
                 depth_multiplier=1.0,
                 prediction_fn=slim.softmax,
                 spatial_squeeze=True,
                 reuse=None,
                 create_aux_logits=True,
                 scope='InceptionV3',
                 global_pool=False):
    """Inception model from http://arxiv.org/abs/1512.00567.
    "Rethinking the Inception Architecture for Computer Vision"
    Christian Szegedy, Vincent Vanhoucke, Sergey Ioffe, Jonathon Shlens,
    Zbigniew Wojna.
    With the default arguments this method constructs the exact model defined in
    the paper. However, one can experiment with variations of the inception_v3
    network by changing arguments dropout_keep_prob, min_depth and
    depth_multiplier.
    The default image size used to train this network is 299x299.
    Args:
      inputs: a tensor of size [batch_size, height, width, channels].
      num_classes: number of predicted classes. If 0 or None, the logits layer
        is omitted and the input features to the logits layer (before dropout)
        are returned instead.
      is_training: whether is training or not.
      dropout_keep_prob: the percentage of activation values that are retained.
      min_depth: Minimum depth value (number of channels) for all convolution ops.
        Enforced when depth_multiplier < 1, and not an active constraint when
        depth_multiplier >= 1.
      depth_multiplier: Float multiplier for the depth (number of channels)
        for all convolution ops. The value must be greater than zero. Typical
        usage will be to set this value in (0, 1) to reduce the number of
        parameters or computation cost of the model.
      prediction_fn: a function to get predictions out of logits.
      spatial_squeeze: if True, logits is of shape [B, C], if false logits is of
          shape [B, 1, 1, C], where B is batch_size and C is number of classes.
      reuse: whether or not the network and its variables should be reused. To be
        able to reuse 'scope' must be given.
      create_aux_logits: Whether to create the auxiliary logits.
      scope: Optional variable_scope.
      global_pool: Optional boolean flag to control the avgpooling before the
        logits layer. If false or unset, pooling is done with a fixed window
        that reduces default-sized inputs to 1x1, while larger inputs lead to
        larger outputs. If true, any input size is pooled down to 1x1.
    Returns:
      net: a Tensor with the logits (pre-softmax activations) if num_classes
        is a non-zero integer, or the non-dropped-out input to the logits layer
        if num_classes is 0 or None.
      end_points: a dictionary from components of the network to the corresponding
        activation.
    Raises:
      ValueError: if 'depth_multiplier' is less than or equal to zero.
    """
    if depth_multiplier <= 0:
        raise ValueError('depth_multiplier is not greater than zero.')
    depth = lambda d: max(int(d * depth_multiplier), min_depth)

    with tf.variable_scope(scope, 'InceptionV3', [inputs], reuse=reuse) as scope:
        with slim.arg_scope([slim.batch_norm, slim.dropout],
                            is_training=is_training):
            net, end_points = inception_v3_base(
                inputs, scope=scope, min_depth=min_depth,
                depth_multiplier=depth_multiplier)

            # Auxiliary Head logits
            if create_aux_logits and num_classes:
                with slim.arg_scope([slim.conv2d, slim.max_pool2d, slim.avg_pool2d],
                                    stride=1, padding='SAME'):
                    aux_logits = end_points['Mixed_6e']
                    with tf.variable_scope('AuxLogits'):
                        aux_logits = slim.avg_pool2d(
                            aux_logits, [5, 5], stride=3, padding='VALID',
                            scope='AvgPool_1a_5x5')
                        aux_logits = slim.conv2d(aux_logits, depth(128), [1, 1],
                                                 scope='Conv2d_1b_1x1')

                        # Shape of feature map before the final layer.
                        kernel_size = _reduced_kernel_size_for_small_input(
                            aux_logits, [5, 5])
                        aux_logits = slim.conv2d(
                            aux_logits, depth(768), kernel_size,
                            weights_initializer=trunc_normal(0.01),
                            padding='VALID', scope='Conv2d_2a_{}x{}'.format(*kernel_size))
                        aux_logits = slim.conv2d(
                            aux_logits, num_classes, [1, 1], activation_fn=None,
                            normalizer_fn=None, weights_initializer=trunc_normal(0.001),
                            scope='Conv2d_2b_1x1')
                        if spatial_squeeze:
                            aux_logits = tf.squeeze(aux_logits, [1, 2], name='SpatialSqueeze')
                        end_points['AuxLogits'] = aux_logits

            # Final pooling and prediction
            with tf.variable_scope('Logits'):
                if global_pool:
                    # Global average pooling.
                    net = tf.reduce_mean(net, [1, 2], keep_dims=True, name='GlobalPool')
                    end_points['global_pool'] = net
                else:
                    # Pooling with a fixed kernel size.
                    kernel_size = _reduced_kernel_size_for_small_input(net, [8, 8])
                    net = slim.avg_pool2d(net, kernel_size, padding='VALID',
                                          scope='AvgPool_1a_{}x{}'.format(*kernel_size))
                    end_points['AvgPool_1a'] = net
                if not num_classes:
                    return net, end_points
                # 1 x 1 x 2048
                net = slim.dropout(net, keep_prob=dropout_keep_prob, scope='Dropout_1b')
                end_points['PreLogits'] = net
                # 2048
                logits = slim.conv2d(net, num_classes, [1, 1], activation_fn=None,
                                     normalizer_fn=None, scope='Conv2d_1c_1x1')
                if spatial_squeeze:
                    logits = tf.squeeze(logits, [1, 2], name='SpatialSqueeze')
                # 1000
            end_points['Logits'] = logits
            end_points['Predictions'] = prediction_fn(logits, scope='Predictions')
    return logits, end_points


inception_v3.default_image_size = 299


def _reduced_kernel_size_for_small_input(input_tensor, kernel_size):
    """Define kernel size which is automatically reduced for small input.
    If the shape of the input images is unknown at graph construction time this
    function assumes that the input images are is large enough.
    Args:
      input_tensor: input tensor of size [batch_size, height, width, channels].
      kernel_size: desired kernel size of length 2: [kernel_height, kernel_width]
    Returns:
      a tensor with the kernel size.
    TODO(jrru): Make this function work with unknown shapes. Theoretically, this
    can be done with the code below. Problems are two-fold: (1) If the shape was
    known, it will be lost. (2) inception.slim.ops._two_element_tuple cannot
    handle tensors that define the kernel size.
        shape = tf.shape(input_tensor)
        return = tf.stack([tf.minimum(shape[1], kernel_size[0]),
                           tf.minimum(shape[2], kernel_size[1])])
    """
    shape = input_tensor.get_shape().as_list()
    if shape[1] is None or shape[2] is None:
        kernel_size_out = kernel_size
    else:
        kernel_size_out = [min(shape[1], kernel_size[0]),
                           min(shape[2], kernel_size[1])]
    return kernel_size_out


inception_v3_arg_scope = inception_utils.inception_arg_scope



import os
from typing import List, Tuple, Any

import numpy as np

from .helper import ListKeyIndexer
from ..base import BaseChunkIndexer as BCI


class AnnoyIndexer(BCI):

    def __init__(self, num_dim: int, data_path: str, metric: str = 'angular', n_trees: int = 10, *args, **kwargs):
        """
        Initialize an AnnoyIndexer

        :param num_dim: when set to -1, then num_dim is auto decided on first .add()
        :param data_path: index data file managed by the annoy indexer
        :param metric:
        :param n_trees:
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.num_dim = num_dim
        self.data_path = data_path
        self.metric = metric
        self.n_trees = n_trees
        self.helper_indexer = ListKeyIndexer()

    def post_init(self):
        from annoy import AnnoyIndex
        self._index = AnnoyIndex(self.num_dim, self.metric) if self.num_dim >= 0 else None
        try:
            if not os.path.exists(self.data_path):
                raise FileNotFoundError('"data_path" is not exist')
            if os.path.isdir(self.data_path):
                raise IsADirectoryError('"data_path" must be a file path, not a directory')
            self._index.load(self.data_path)
        except:
            self.logger.warning('fail to load model from %s, will create an empty one' % self.data_path)

    @BCI.update_helper_indexer
    def add(self, keys: List[Tuple[int, Any]], vectors: np.ndarray, weights: List[float], *args, **kwargs):
        last_idx = self.helper_indexer.num_chunks

        if len(vectors) != len(keys):
            raise ValueError('vectors length should be equal to doc_ids')

        if vectors.dtype != np.float32:
            raise ValueError("vectors should be ndarray of float32")

        if self._index is None:
            from annoy import AnnoyIndex
            # means num_dim in unknown during init
            self.num_dim = vectors.shape[1]
            self._index = AnnoyIndex(self.num_dim, self.metric)

        for idx, vec in enumerate(vectors):
            self._index.add_item(last_idx + idx, vec)

    def query(self, keys: 'np.ndarray', top_k: int, *args, **kwargs) -> List[List[Tuple]]:
        self._index.build(self.n_trees)
        if keys.dtype != np.float32:
            raise ValueError('vectors should be ndarray of float32')
        res = []
        for k in keys:
            ret, relevance_score = self._index.get_nns_by_vector(k, top_k, include_distances=True)
            chunk_info = self.helper_indexer.query(ret)
            res.append([(*r, s) for r, s in zip(chunk_info, relevance_score)])
        return res

    def __getstate__(self):
        d = super().__getstate__()
        self._index.save(self.data_path)
        return d



import pickle
from typing import List

import gc

from ..base import BaseDocIndexer as BDI
from ...proto import gnes_pb2


class RocksDBIndexer(BDI):

    def __init__(self, data_path: str,
                 drop_raw_data: bool = False,
                 drop_chunk_blob: bool = False,
                 read_only: bool = False,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_path = data_path
        self.drop_raw_data = drop_raw_data
        self.drop_chunk_blob = drop_chunk_blob
        self.read_only = read_only
        self.kwargs = kwargs

    def post_init(self):
        import rocksdb

        opts = rocksdb.Options()
        opts.create_if_missing = True
        opts.max_open_files = 300000
        opts.write_buffer_size = 67108864
        opts.max_write_buffer_number = 3
        opts.target_file_size_base = 67108864

        opts.table_factory = rocksdb.BlockBasedTableFactory(
            filter_policy=rocksdb.BloomFilterPolicy(10),
            block_cache=rocksdb.LRUCache(2 * (1024 ** 3)),
            block_cache_compressed=rocksdb.LRUCache(500 * (1024 ** 2)))

        for key, value in self.kwargs.items():
            setattr(opts, key, value)

        self._db = rocksdb.DB(self.data_path, opts, read_only=self.read_only)

    @BDI.update_counter
    def add(self, keys: List[int], docs: List['gnes_pb2.Document'], *args, **kwargs):
        import rocksdb

        write_batch = rocksdb.WriteBatch()
        for k, d in zip(keys, docs):
            key_bytes = pickle.dumps(k)
            if self.drop_raw_data:
                d.ClearField('raw_data')
            if self.drop_chunk_blob:
                for c in d.chunks:
                    c.ClearField('blob')
            value_bytes = d.SerializeToString()
            write_batch.put(key_bytes, value_bytes)
        self._db.write(write_batch, sync=True)

    def query(self, keys: List[int], *args, **kwargs) -> List['gnes_pb2.Document']:
        query_keys = []
        for k in keys:
            key_value = pickle.dumps(k)
            query_keys.append(key_value)

        values = self._db.multi_get(query_keys)

        docs = []
        for k in query_keys:
            v = values[k]
            if v is not None:
                _doc = gnes_pb2.Document()
                _doc.ParseFromString(v)
                docs.append(_doc)
            else:
                docs.append(None)
        return docs

    def scan(self, reversed_scan: bool = False):
        iterator = self._db.iterkeys()

        if reversed_scan:
            iterator.seek_to_last()
        else:
            iterator.seek_to_first()

        if reversed_scan:
            iterator = reversed(iterator)

        for key_bytes in iterator:
            doc_id = pickle.loads(key_bytes)
            value_bytes = self._db.get(key_bytes)
            pb_doc = gnes_pb2.Document()
            pb_doc.ParseFromString(value_bytes)

            yield doc_id, pb_doc

    def close(self):
        super().close()
        try:
            del self._db
        except AttributeError:
            pass
        gc.collect()
